@@ -21,8 +21,10 @@ THE SOFTWARE.
 """
 
 from sys import intern
+from warnings import warn
+from typing import Any, Dict, List, Optional, Sequence, Tuple, Union
 
-from pyopencl.version import VERSION, VERSION_STATUS, VERSION_TEXT  # noqa
+from pyopencl.version import VERSION, VERSION_STATUS, VERSION_TEXT  # noqa: F401
 
 # must import, otherwise dtype registry will not be fully populated
 import pyopencl.cltypes  # noqa: F401
@@ -40,10 +42,8 @@ os.environ["PYOPENCL_HOME"] = os.path.dirname(os.path.abspath(__file__))
 try:
     import pyopencl._cl as _cl
 except ImportError:
-    import os
     from os.path import dirname, join, realpath
     if realpath(join(os.getcwd(), "pyopencl")) == realpath(dirname(__file__)):
-        from warnings import warn
         warn("It looks like you are importing PyOpenCL from "
                 "its source directory. This likely won't work.")
     raise
@@ -173,17 +173,17 @@ except ImportError:
 
 if not _PYPY:
     # FIXME: Add back to default set when pypy support catches up
-    from pyopencl._cl import (  # noqa
+    from pyopencl._cl import (  # noqa: F401
         enqueue_map_buffer,
         enqueue_map_image,
         )
 
 if get_cl_header_version() >= (1, 1):
-    from pyopencl._cl import (  # noqa
+    from pyopencl._cl import (  # noqa: F401
         UserEvent,
         )
 if get_cl_header_version() >= (1, 2):
-    from pyopencl._cl import (  # noqa
+    from pyopencl._cl import (  # noqa: F401
         _enqueue_marker_with_wait_list,
         _enqueue_barrier_with_wait_list,
 
@@ -198,16 +198,14 @@ if get_cl_header_version() >= (1, 2):
         )
 
 if get_cl_header_version() >= (2, 0):
-    from pyopencl._cl import (  # noqa
-        SVMAllocation,
+    from pyopencl._cl import (  # noqa: F401
+        SVMPointer,
         SVM,
-
-        # FIXME
-        #enqueue_svm_migratemem,
+        SVMAllocation,
         )
 
 if _cl.have_gl():
-    from pyopencl._cl import (  # noqa
+    from pyopencl._cl import (  # noqa: F401
         gl_object_type,
         gl_texture_info,
 
@@ -217,12 +215,12 @@ if _cl.have_gl():
         )
 
     try:
-        from pyopencl._cl import get_apple_cgl_share_group  # noqa
+        from pyopencl._cl import get_apple_cgl_share_group  # noqa: F401
     except ImportError:
         pass
 
     try:
-        from pyopencl._cl import (  # noqa
+        from pyopencl._cl import (  # noqa: F401
             enqueue_acquire_gl_objects,
             enqueue_release_gl_objects,
         )
@@ -266,8 +264,6 @@ class CommandQueueUsedAfterExit(UserWarning):
 
 
 def compiler_output(text):
-    import os
-    from warnings import warn
     if int(os.environ.get("PYOPENCL_COMPILER_OUTPUT", "0")):
         warn(text, CompilerWarning)
     else:
@@ -363,11 +359,13 @@ def _options_to_bytestring(options):
 
 # {{{ Program (wrapper around _Program, adds caching support)
 
-_DEFAULT_BUILD_OPTIONS = []
-_DEFAULT_INCLUDE_OPTIONS = ["-I", _find_pyopencl_include_path()]
+_PYOPENCL_NO_CACHE: bool = "PYOPENCL_NO_CACHE" in os.environ
+
+_DEFAULT_BUILD_OPTIONS: List[str] = []
+_DEFAULT_INCLUDE_OPTIONS: List[str] = ["-I", _find_pyopencl_include_path()]
 
 # map of platform.name to build options list
-_PLAT_BUILD_OPTIONS = {
+_PLAT_BUILD_OPTIONS: Dict[str, List[str]] = {
         "Oclgrind": ["-D", "PYOPENCL_USING_OCLGRIND"],
         }
 
@@ -386,10 +384,8 @@ def enable_debugging(platform_or_context):
     if "AMD Accelerated" in platform.name:
         _PLAT_BUILD_OPTIONS.setdefault(platform.name, []).extend(
                 ["-g", "-O0"])
-        import os
         os.environ["CPU_MAX_COMPUTE_UNITS"] = "1"
     else:
-        from warnings import warn
         warn("do not know how to enable debugging on '%s'"
                 % platform.name)
 
@@ -428,7 +424,6 @@ class Program:
             return self._prg
         else:
             # "no program" can only happen in from-source case.
-            from warnings import warn
             warn("Pre-build attribute access defeats compiler caching.",
                     stacklevel=3)
 
@@ -447,14 +442,15 @@ class Program:
             knl._setup(self)
         return result
 
+    @property
     def int_ptr(self):
         return self._get_prg().int_ptr
-    int_ptr = property(int_ptr, doc=_cl._Program.int_ptr.__doc__)
+    int_ptr.__doc__ = _cl._Program.int_ptr.__doc__
 
+    @staticmethod
     def from_int_ptr(int_ptr_value, retain=True):
         return Program(_cl._Program.from_int_ptr(int_ptr_value, retain))
     from_int_ptr.__doc__ = _cl._Program.from_int_ptr.__doc__
-    from_int_ptr = staticmethod(from_int_ptr)
 
     def __getattr__(self, attr):
         try:
@@ -495,7 +491,6 @@ class Program:
                 + _PLAT_BUILD_OPTIONS.get(
                     context.devices[0].platform.name, []))
 
-        import os
         forced_options = os.environ.get("PYOPENCL_BUILD_OPTIONS")
         if forced_options:
             options = options + forced_options.split()
@@ -511,10 +506,8 @@ class Program:
         if cache_dir is None:
             cache_dir = getattr(self._context, "cache_dir", None)
 
-        import os
         build_descr = None
-
-        if os.environ.get("PYOPENCL_NO_CACHE") and self._prg is None:
+        if _PYOPENCL_NO_CACHE and self._prg is None:
             build_descr = "uncached source build (cache disabled by user)"
             self._prg = _cl._Program(self._context, self._source)
 
@@ -662,7 +655,6 @@ def _add_functionality():
         return ("v1", self.vendor, self.vendor_id, self.name, self.version)
 
     def device_persistent_unique_id(self):
-        from warnings import warn
         warn("Device.persistent_unique_id is deprecated. "
                 "Use Device.hashable_model_and_version_identifier instead.",
                 DeprecationWarning, stacklevel=2)
@@ -684,7 +676,6 @@ def _add_functionality():
 
     def context_init(self, devices, properties, dev_type, cache_dir=None):
         if cache_dir is not None:
-            from warnings import warn
             warn("The 'cache_dir' argument to the Context constructor "
                 "is deprecated and no longer has an effect. "
                 "It was removed because it only applied to the wrapper "
@@ -875,7 +866,13 @@ def _add_functionality():
             __call__ = enqueue
             set_args = my_set_args
 
-        self.__class__ = KernelWithCustomEnqueue
+        try:
+            self.__class__ = KernelWithCustomEnqueue
+        except TypeError:
+            # __class__ assignment may not work in all cases, due to differing
+            # object layouts. Fall back to bouncing through kernel_call below.
+            self._enqueue = enqueue
+            self.set_args = my_set_args
 
     def kernel_get_work_group_info(self, param, device):
         cache_key = (param, device.int_ptr)
@@ -900,10 +897,10 @@ def _add_functionality():
         # kernel_set_scalar_arg_dtypes is called.
         return self._enqueue(self, queue, global_size, local_size, *args, **kwargs)
 
-    def kernel_capture_call(self, filename, queue, global_size, local_size,
+    def kernel_capture_call(self, output_file, queue, global_size, local_size,
             *args, **kwargs):
         from pyopencl.capture_call import capture_kernel_call
-        capture_kernel_call(self, filename, queue, global_size, local_size,
+        capture_kernel_call(self, output_file, queue, global_size, local_size,
                 *args, **kwargs)
 
     def kernel_get_info(self, param_name):
@@ -970,7 +967,6 @@ def _add_functionality():
 
         if hostbuf is not None and not \
                 (flags & (mem_flags.USE_HOST_PTR | mem_flags.COPY_HOST_PTR)):
-            from warnings import warn
             warn("'hostbuf' was passed, but no memory flags to make use of it.")
 
         if hostbuf is None and pitches is not None:
@@ -1043,7 +1039,6 @@ def _add_functionality():
 
     class _ImageInfoGetter:
         def __init__(self, event):
-            from warnings import warn
             warn("Image.image.attr is deprecated and will go away in 2021. "
                     "Use Image.attr directly, instead.")
 
@@ -1131,44 +1126,167 @@ def _add_functionality():
 
     # }}}
 
+    # {{{ SVMPointer
+
+    if get_cl_header_version() >= (2, 0):
+        SVMPointer.__doc__ = """A base class for things that can be passed to
+            functions that allow an SVM pointer, e.g. kernel enqueues and memory
+            copies.
+
+            Objects of this type cannot currently be directly created or
+            implemented in Python.  To obtain objects implementing this type,
+            consider its subtypes :class:`SVMAllocation` and :class:`SVM`.
+
+
+            .. property:: svm_ptr
+
+                Gives the SVM pointer as an :class:`int`.
+
+            .. property:: size
+
+                An :class:`int` denoting the size in bytes, or *None*, if the size
+                of the SVM pointed to is not known.
+
+                *Most* objects of this type (e.g. instances of
+                :class:`SVMAllocation` and :class:`SVM` know their size, so that,
+                for example :class:`enqueue_copy` will automatically copy an entire
+                :class:`SVMAllocation` when a size is not explicitly specified.
+
+            .. automethod:: map
+            .. automethod:: map_ro
+            .. automethod:: map_rw
+            .. automethod:: as_buffer
+            .. property:: buf
+
+                An opaque object implementing the :c:func:`Python buffer protocol
+                <PyObject_GetBuffer>`. It exposes the pointed-to memory as
+                a one-dimensional buffer of bytes, with the size matching
+                :attr:`size`.
+
+                No guarantee is provided that two references to this attribute
+                result in the same object.
+            """
+
+    def svmptr_map(self, queue: CommandQueue, *, flags: int, is_blocking: bool =
+                   True, wait_for: Optional[Sequence[Event]] = None,
+                   size: Optional[Event] = None) -> "SVMMap":
+        """
+        :arg is_blocking: If *False*, subsequent code must wait on
+            :attr:`SVMMap.event` in the returned object before accessing the
+            mapped memory.
+        :arg flags: a combination of :class:`pyopencl.map_flags`.
+        :arg size: The size of the map in bytes. If not provided, defaults to
+            :attr:`size`.
+
+        |std-enqueue-blurb|
+        """
+        return SVMMap(self,
+                np.asarray(self.buf),
+                queue,
+                _cl._enqueue_svm_map(queue, is_blocking, flags, self, wait_for,
+                                    size=size))
+
+    def svmptr_map_ro(self, queue: CommandQueue, *, is_blocking: bool = True,
+                      wait_for: Optional[Sequence[Event]] = None,
+                      size: Optional[int] = None) -> "SVMMap":
+        """Like :meth:`map`, but with *flags* set for a read-only map.
+        """
+
+        return self.map(queue, flags=map_flags.READ,
+                is_blocking=is_blocking, wait_for=wait_for, size=size)
+
+    def svmptr_map_rw(self, queue: CommandQueue, *, is_blocking: bool = True,
+                      wait_for: Optional[Sequence[Event]] = None,
+                      size: Optional[int] = None) -> "SVMMap":
+        """Like :meth:`map`, but with *flags* set for a read-only map.
+        """
+
+        return self.map(queue, flags=map_flags.READ | map_flags.WRITE,
+                is_blocking=is_blocking, wait_for=wait_for, size=size)
+
+    def svmptr__enqueue_unmap(self, queue, wait_for=None):
+        return _cl._enqueue_svm_unmap(queue, self, wait_for)
+
+    def svmptr_as_buffer(self, ctx: Context, *, flags: Optional[int] = None,
+                         size: Optional[int] = None) -> Buffer:
+        """
+        :arg ctx: a :class:`Context`
+        :arg flags: a combination of :class:`pyopencl.map_flags`, defaults to
+            read-write.
+        :arg size: The size of the map in bytes. If not provided, defaults to
+            :attr:`size`.
+        :returns: a :class:`Buffer` corresponding to *self*.
+
+        The memory referred to by this object must not be freed before
+        the returned :class:`Buffer` is released.
+        """
+
+        if flags is None:
+            flags = mem_flags.READ_WRITE | mem_flags.USE_HOST_PTR
+
+        if size is None:
+            size = self.size
+
+        return Buffer(ctx, flags, size=size, hostbuf=self.buf)
+
+    if get_cl_header_version() >= (2, 0):
+        SVMPointer.map = svmptr_map
+        SVMPointer.map_ro = svmptr_map_ro
+        SVMPointer.map_rw = svmptr_map_rw
+        SVMPointer._enqueue_unmap = svmptr__enqueue_unmap
+        SVMPointer.as_buffer = svmptr_as_buffer
+
+    # }}}
+
     # {{{ SVMAllocation
 
     if get_cl_header_version() >= (2, 0):
-        SVMAllocation.__doc__ = """An object whose lifetime is tied to an allocation of shared virtual memory.
-
-            .. note::
-
-                Most likely, you will not want to use this directly, but rather
-                :func:`svm_empty` and related functions which allow access to this
-                functionality using a friendlier, more Pythonic interface.
+        SVMAllocation.__doc__ = """
+            Is a :class:`SVMPointer`.
 
             .. versionadded:: 2016.2
 
-            .. automethod:: __init__(self, ctx, size, alignment, flags=None)
-            .. automethod:: release
+            .. automethod:: __init__
+
+                :arg flags: See :class:`svm_mem_flags`.
+                :arg queue: If not specified, the allocation will be freed
+                    eagerly, irrespective of whether pending/enqueued operations
+                    are still using this memory.
+
+                    If specified, deallocation of the memory will be enqueued
+                    with the given queue, and will only be performed
+                    after previously-enqueue operations in the queue have
+                    completed.
+
+                    It is an error to specify an out-of-order queue.
+
+                    .. warning::
+
+                        Not specifying a queue will typically lead to undesired
+                        behavior, including crashes and memory corruption.
+                        See the warning in :ref:`svm`.
+
             .. automethod:: enqueue_release
+
+                Enqueue the release of this allocation into *queue*.
+                If *queue* is not specified, enqueue the deallocation
+                into the queue provided at allocation time or via
+                :class:`bind_to_queue`.
+
+            .. automethod:: bind_to_queue
+
+                Change the queue used for implicit enqueue of deallocation
+                to *queue*. Sufficient synchronization is ensured by
+                enqueuing a marker into the old queue and waiting on this
+                marker in the new queue.
+
+            .. automethod:: unbind_from_queue
+
+                Configure the allocation to no longer implicitly enqueue
+                memory allocation. If such a queue was previously provided,
+                :meth:`~CommandQueue.finish` is automatically called on it.
             """
 
-    if get_cl_header_version() >= (2, 0):
-        svmallocation_old_init = SVMAllocation.__init__
-
-    def svmallocation_init(self, ctx, size, alignment, flags, _interface=None):
-        """
-        :arg ctx: a :class:`Context`
-        :arg flags: some of :class:`svm_mem_flags`.
-        """
-        svmallocation_old_init(self, ctx, size, alignment, flags)
-
-        # mem_flags.READ_ONLY applies to kernels, not the host
-        read_write = True
-        _interface["data"] = (
-                int(self._ptr_as_int()), not read_write)
-
-        self.__array_interface__ = _interface
-
-    if get_cl_header_version() >= (2, 0):
-        SVMAllocation.__init__ = svmallocation_init
-
     # }}}
 
     # {{{ SVM
@@ -1178,23 +1296,14 @@ def _add_functionality():
             (such as a :class:`numpy.ndarray`) as referring to shared virtual
             memory.
 
+            Is a :class:`SVMPointer`, hence objects of this type may be passed
+            to kernel calls and :func:`enqueue_copy`, and all methods declared
+            there are also available there. Note that :meth:`map` differs
+            slightly from :meth:`SVMPointer.map`.
+
             Depending on the features of the OpenCL implementation, the following
             types of objects may be passed to/wrapped in this type:
 
-            *   coarse-grain shared memory as returned by (e.g.) :func:`csvm_empty`
-                for any implementation of OpenCL 2.0.
-
-                This is how coarse-grain SVM may be used from both host and device::
-
-                    svm_ary = cl.SVM(
-                        cl.csvm_empty(ctx, 1000, np.float32, alignment=64))
-                    assert isinstance(svm_ary.mem, np.ndarray)
-
-                    with svm_ary.map_rw(queue) as ary:
-                        ary.fill(17)  # use from host
-
-                    prg.twice(queue, svm_ary.mem.shape, None, svm_ary)
-
             *   fine-grain shared memory as returned by (e.g.) :func:`fsvm_empty`,
                 if the implementation supports fine-grained shared virtual memory.
                 This memory may directly be passed to a kernel::
@@ -1221,10 +1330,28 @@ def _add_functionality():
                     queue.finish() # synchronize
                     print(ary) # access from host
 
-            Objects of this type may be passed to kernel calls and
-            :func:`enqueue_copy`.  Coarse-grain shared-memory *must* be mapped
-            into host address space using :meth:`map` before being accessed
-            through the :mod:`numpy` interface.
+            *   coarse-grain shared memory as returned by (e.g.) :func:`csvm_empty`
+                for any implementation of OpenCL 2.0.
+
+                .. note::
+
+                    Applications making use of coarse-grain SVM may be better
+                    served by opaque-style SVM. See :ref:`opaque-svm`.
+
+                This is how coarse-grain SVM may be used from both host and device::
+
+                    svm_ary = cl.SVM(
+                        cl.csvm_empty(ctx, 1000, np.float32, alignment=64))
+                    assert isinstance(svm_ary.mem, np.ndarray)
+
+                    with svm_ary.map_rw(queue) as ary:
+                        ary.fill(17)  # use from host
+
+                    prg.twice(queue, svm_ary.mem.shape, None, svm_ary)
+
+            Coarse-grain shared-memory *must* be mapped into host address space
+            using :meth:`~SVMPointer.map` before being accessed through the
+            :mod:`numpy` interface.
 
             .. note::
 
@@ -1245,9 +1372,10 @@ def _add_functionality():
             .. automethod:: map
             .. automethod:: map_ro
             .. automethod:: map_rw
-            .. automethod:: as_buffer
             """
 
+    # }}}
+
     if get_cl_header_version() >= (2, 0):
         svm_old_init = SVM.__init__
 
@@ -1261,14 +1389,18 @@ def _add_functionality():
         :arg is_blocking: If *False*, subsequent code must wait on
             :attr:`SVMMap.event` in the returned object before accessing the
             mapped memory.
-        :arg flags: a combination of :class:`pyopencl.map_flags`, defaults to
-            read-write.
+        :arg flags: a combination of :class:`pyopencl.map_flags`.
         :returns: an :class:`SVMMap` instance
 
+        This differs from the inherited :class:`SVMPointer.map` in that no size
+        can be specified, and that :attr:`mem` is the exact array produced
+        when the :class:`SVMMap` is used as a context manager.
+
         |std-enqueue-blurb|
         """
         return SVMMap(
                 self,
+                self.mem,
                 queue,
                 _cl._enqueue_svm_map(queue, is_blocking, flags, self, wait_for))
 
@@ -1287,29 +1419,12 @@ def _add_functionality():
     def svm__enqueue_unmap(self, queue, wait_for=None):
         return _cl._enqueue_svm_unmap(queue, self, wait_for)
 
-    def svm_as_buffer(self, ctx, flags=None):
-        """
-        :arg ctx: a :class:`Context`
-        :arg flags: a combination of :class:`pyopencl.map_flags`, defaults to
-            read-write.
-        :returns: a :class:`Buffer` corresponding to *self*.
-
-        The memory referred to by this object must not be freed before
-        the returned :class:`Buffer` is released.
-        """
-
-        if flags is None:
-            flags = mem_flags.READ_WRITE
-
-        return Buffer(ctx, flags, size=self.mem.nbytes, hostbuf=self.mem)
-
     if get_cl_header_version() >= (2, 0):
         SVM.__init__ = svm_init
         SVM.map = svm_map
         SVM.map_ro = svm_map_ro
         SVM.map_rw = svm_map_rw
         SVM._enqueue_unmap = svm__enqueue_unmap
-        SVM.as_buffer = svm_as_buffer
 
     # }}}
 
@@ -1412,10 +1527,30 @@ _add_functionality()
 # }}}
 
 
+# {{{ _OverriddenArrayInterfaceSVMAllocation
+
+if get_cl_header_version() >= (2, 0):
+    class _OverriddenArrayInterfaceSVMAllocation(SVMAllocation):
+        def __init__(self, ctx, size, alignment, flags, *, _interface,
+                queue=None):
+            """
+            :arg ctx: a :class:`Context`
+            :arg flags: some of :class:`svm_mem_flags`.
+            """
+            super().__init__(ctx, size, alignment, flags, queue)
+
+            # mem_flags.READ_ONLY applies to kernels, not the host
+            read_write = True
+            _interface["data"] = (int(self.svm_ptr), not read_write)
+
+            self.__array_interface__ = _interface
+
+# }}}
+
+
 # {{{ create_some_context
 
 def create_some_context(interactive=None, answers=None):
-    import os
     if answers is None:
         if "PYOPENCL_CTX" in os.environ:
             ctx_spec = os.environ["PYOPENCL_CTX"]
@@ -1438,7 +1573,6 @@ def create_some_context(interactive=None, answers=None):
     if interactive is None:
         interactive = True
         try:
-            import sys
             if not sys.stdin.isatty():
                 interactive = False
         except Exception:
@@ -1552,19 +1686,24 @@ _csc = create_some_context
 
 class SVMMap:
     """
-    .. attribute:: event
-
-    .. versionadded:: 2016.2
-
-    .. automethod:: release
-
+    Returned by :func:`SVMPointer.map` and :func:`SVM.map`.
     This class may also be used as a context manager in a ``with`` statement.
     :meth:`release` will be called upon exit from the ``with`` region.
     The value returned to the ``as`` part of the context manager is the
     mapped Python object (e.g. a :mod:`numpy` array).
+
+    .. versionadded:: 2016.2
+
+    .. property:: event
+
+        The :class:`Event` returned when mapping the memory.
+
+    .. automethod:: release
+
     """
-    def __init__(self, svm, queue, event):
+    def __init__(self, svm, array, queue, event):
         self.svm = svm
+        self.array = array
         self.queue = queue
         self.event = event
 
@@ -1573,7 +1712,7 @@ class SVMMap:
             self.release()
 
     def __enter__(self):
-        return self.svm.mem
+        return self.array
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         self.release()
@@ -1597,6 +1736,11 @@ class SVMMap:
 
 # {{{ enqueue_copy
 
+_IMAGE_MEM_OBJ_TYPES = [mem_object_type.IMAGE2D, mem_object_type.IMAGE3D]
+if get_cl_header_version() >= (1, 2):
+    _IMAGE_MEM_OBJ_TYPES.append(mem_object_type.IMAGE2D_ARRAY)
+
+
 def enqueue_copy(queue, dest, src, **kwargs):
     """Copy from :class:`Image`, :class:`Buffer` or the host to
     :class:`Image`, :class:`Buffer` or the host. (Note: host-to-host
@@ -1623,7 +1767,7 @@ def enqueue_copy(queue, dest, src, **kwargs):
         Two types of 'buffer' occur in the arguments to this function,
         :class:`Buffer` and 'host-side buffers'. The latter are
         defined by Python and commonly called `buffer objects
-        <https://docs.python.org/3.4/c-api/buffer.html>`_. :mod:`numpy`
+        <https://docs.python.org/3/c-api/buffer.html>`__. :mod:`numpy`
         arrays are a very common example.
         Make sure to always be clear on whether a :class:`Buffer` or a
         Python buffer object is needed.
@@ -1632,7 +1776,13 @@ def enqueue_copy(queue, dest, src, **kwargs):
     .. rubric :: Transfer :class:`Buffer` ↔ host
     .. ------------------------------------------------------------------------
 
-    :arg device_offset: offset in bytes (optional)
+    :arg src_offset: offset in bytes (optional)
+
+        May only be nonzero if applied on the device side.
+
+    :arg dst_offset: offset in bytes (optional)
+
+        May only be nonzero if applied on the device side.
 
     .. note::
 
@@ -1652,7 +1802,7 @@ def enqueue_copy(queue, dest, src, **kwargs):
         and to the minimum of the size of the source and target
         from 2013.1 on.
     :arg src_offset: (optional)
-    :arg dest_offset: (optional)
+    :arg dst_offset: (optional)
 
     .. ------------------------------------------------------------------------
     .. rubric :: Rectangular :class:`Buffer` ↔  host transfers (CL 1.1 and newer)
@@ -1718,7 +1868,7 @@ def enqueue_copy(queue, dest, src, **kwargs):
         three or shorter. (mandatory)
 
     .. ------------------------------------------------------------------------
-    .. rubric :: Transfer :class:`SVM`/host ↔ :class:`SVM`/host
+    .. rubric :: Transfer :class:`SVMPointer`/host ↔ :class:`SVMPointer`/host
     .. ------------------------------------------------------------------------
 
     :arg byte_count: (optional) If not specified, defaults to the
@@ -1735,30 +1885,66 @@ def enqueue_copy(queue, dest, src, **kwargs):
         if dest.type == mem_object_type.BUFFER:
             if isinstance(src, MemoryObjectHolder):
                 if src.type == mem_object_type.BUFFER:
+                    # {{{ buffer -> buffer
+
                     if "src_origin" in kwargs:
+                        # rectangular
                         return _cl._enqueue_copy_buffer_rect(
                                 queue, src, dest, **kwargs)
                     else:
-                        kwargs["dst_offset"] = kwargs.pop("dest_offset", 0)
+                        # linear
+                        dest_offset = kwargs.pop("dest_offset", None)
+                        if dest_offset is not None:
+                            if "dst_offset" in kwargs:
+                                raise TypeError("may not specify both 'dst_offset' "
+                                                "and 'dest_offset'")
+
+                            warn("The 'dest_offset' argument of enqueue_copy "
+                                 "is deprecated. Use 'dst_offset' instead. "
+                                 "'dest_offset' will stop working in 2023.x.",
+                                 DeprecationWarning, stacklevel=2)
+
+                            kwargs["dst_offset"] = dest_offset
+
                         return _cl._enqueue_copy_buffer(queue, src, dest, **kwargs)
-                elif src.type in [mem_object_type.IMAGE2D, mem_object_type.IMAGE3D]:
+
+                    # }}}
+                elif src.type in _IMAGE_MEM_OBJ_TYPES:
                     return _cl._enqueue_copy_image_to_buffer(
                             queue, src, dest, **kwargs)
                 else:
                     raise ValueError("invalid src mem object type")
             else:
-                # assume from-host
+                # {{{ host -> buffer
+
                 if "buffer_origin" in kwargs:
                     return _cl._enqueue_write_buffer_rect(queue, dest, src, **kwargs)
                 else:
+                    device_offset = kwargs.pop("device_offset", None)
+                    if device_offset is not None:
+                        if "dst_offset" in kwargs:
+                            raise TypeError("may not specify both 'device_offset' "
+                                            "and 'dst_offset'")
+
+                        warn("The 'device_offset' argument of enqueue_copy "
+                                "is deprecated. Use 'dst_offset' instead. "
+                                "'dst_offset' will stop working in 2023.x.",
+                                DeprecationWarning, stacklevel=2)
+
+                        kwargs["dst_offset"] = device_offset
+
                     return _cl._enqueue_write_buffer(queue, dest, src, **kwargs)
 
-        elif dest.type in [mem_object_type.IMAGE2D, mem_object_type.IMAGE3D]:
+                # }}}
+
+        elif dest.type in _IMAGE_MEM_OBJ_TYPES:
+            # {{{ ... -> image
+
             if isinstance(src, MemoryObjectHolder):
                 if src.type == mem_object_type.BUFFER:
                     return _cl._enqueue_copy_buffer_to_image(
                             queue, src, dest, **kwargs)
-                elif src.type in [mem_object_type.IMAGE2D, mem_object_type.IMAGE3D]:
+                elif src.type in _IMAGE_MEM_OBJ_TYPES:
                     return _cl._enqueue_copy_image(queue, src, dest, **kwargs)
                 else:
                     raise ValueError("invalid src mem object type")
@@ -1775,17 +1961,30 @@ def enqueue_copy(queue, dest, src, **kwargs):
 
                 return _cl._enqueue_write_image(
                         queue, dest, origin, region, src, **kwargs)
+
+            # }}}
         else:
             raise ValueError("invalid dest mem object type")
 
-    elif get_cl_header_version() >= (2, 0) and isinstance(dest, SVM):
-        # to SVM
-        if not isinstance(src, SVM):
+    elif get_cl_header_version() >= (2, 0) and isinstance(dest, SVMPointer):
+        # {{{ ... ->  SVM
+
+        if not isinstance(src, SVMPointer):
             src = SVM(src)
 
         is_blocking = kwargs.pop("is_blocking", True)
+
+        # These are NOT documented. They only support consistency with the
+        # Buffer-based API for the sake of the Array.
+        if kwargs.pop("src_offset", 0) != 0:
+            raise ValueError("src_offset must be 0")
+        if kwargs.pop("dst_offset", 0) != 0:
+            raise ValueError("dst_offset must be 0")
+
         return _cl._enqueue_svm_memcpy(queue, is_blocking, dest, src, **kwargs)
 
+        # }}}
+
     else:
         # assume to-host
 
@@ -1794,8 +1993,22 @@ def enqueue_copy(queue, dest, src, **kwargs):
                 if "buffer_origin" in kwargs:
                     return _cl._enqueue_read_buffer_rect(queue, src, dest, **kwargs)
                 else:
+                    device_offset = kwargs.pop("device_offset", None)
+                    if device_offset is not None:
+                        if "src_offset" in kwargs:
+                            raise TypeError("may not specify both 'device_offset' "
+                                            "and 'src_offset'")
+
+                        warn("The 'device_offset' argument of enqueue_copy "
+                                "is deprecated. Use 'src_offset' instead. "
+                                "'dst_offset' will stop working in 2023.x.",
+                                DeprecationWarning, stacklevel=2)
+
+                        kwargs["src_offset"] = device_offset
+
                     return _cl._enqueue_read_buffer(queue, src, dest, **kwargs)
-            elif src.type in [mem_object_type.IMAGE2D, mem_object_type.IMAGE3D]:
+
+            elif src.type in _IMAGE_MEM_OBJ_TYPES:
                 origin = kwargs.pop("origin")
                 region = kwargs.pop("region")
 
@@ -1809,12 +2022,21 @@ def enqueue_copy(queue, dest, src, **kwargs):
                         queue, src, origin, region, dest, **kwargs)
             else:
                 raise ValueError("invalid src mem object type")
-        elif isinstance(src, SVM):
-            # from svm
+        elif isinstance(src, SVMPointer):
+            # {{{ svm -> host
+
             # dest is not a SVM instance, otherwise we'd be in the branch above
+
+            # This is NOT documented. They only support consistency with the
+            # Buffer-based API for the sake of the Array.
+            if kwargs.pop("src_offset", 0) != 0:
+                raise ValueError("src_offset must be 0")
+
             is_blocking = kwargs.pop("is_blocking", True)
             return _cl._enqueue_svm_memcpy(
                     queue, is_blocking, SVM(dest), src, **kwargs)
+
+            # }}}
         else:
             # assume from-host
             raise TypeError("enqueue_copy cannot perform host-to-host transfers")
@@ -1822,6 +2044,28 @@ def enqueue_copy(queue, dest, src, **kwargs):
 # }}}
 
 
+# {{{ enqueue_fill
+
+def enqueue_fill(queue: CommandQueue,
+        dest: "Union[MemoryObject, SVMPointer]",
+        pattern: Any, size: int, *, offset: int = 0,
+        wait_for: Optional[Sequence[Event]] = None) -> Event:
+    """
+    .. versionadded:: 2022.2
+    """
+    if isinstance(dest, MemoryObjectHolder):
+        return enqueue_fill_buffer(queue, dest, pattern, offset, size, wait_for)
+    elif isinstance(dest, SVMPointer):
+        if offset:
+            raise NotImplementedError("enqueue_fill with SVM does not yet support "
+                    "offsets")
+        return enqueue_svm_memfill(queue, dest, pattern, size, wait_for)
+    else:
+        raise TypeError(f"enqueue_fill does not know how to fill '{type(dest)}'")
+
+# }}}
+
+
 # {{{ image creation
 
 DTYPE_TO_CHANNEL_TYPE = {
@@ -1927,7 +2171,6 @@ def enqueue_barrier(queue, wait_for=None):
 
 def enqueue_fill_buffer(queue, mem, pattern, offset, size, wait_for=None):
     if not (queue._get_cl_version() >= (1, 2) and get_cl_header_version() >= (1, 2)):
-        from warnings import warn
         warn("The context for this queue does not declare OpenCL 1.2 support, so "
                 "the next thing you might see is a crash")
 
@@ -1944,7 +2187,7 @@ def enqueue_fill_buffer(queue, mem, pattern, offset, size, wait_for=None):
 def enqueue_svm_memfill(queue, dest, pattern, byte_count=None, wait_for=None):
     """Fill shared virtual memory with a pattern.
 
-    :arg dest: a Python buffer object, optionally wrapped in an :class:`SVM` object
+    :arg dest: a Python buffer object, or any implementation of :class:`SVMPointer`.
     :arg pattern: a Python buffer object (e.g. a :class:`numpy.ndarray` with the
         fill pattern to be used.
     :arg byte_count: The size of the memory to be fill. Defaults to the
@@ -1955,17 +2198,17 @@ def enqueue_svm_memfill(queue, dest, pattern, byte_count=None, wait_for=None):
     .. versionadded:: 2016.2
     """
 
-    if not isinstance(dest, SVM):
+    if not isinstance(dest, SVMPointer):
         dest = SVM(dest)
 
     return _cl._enqueue_svm_memfill(
-            queue, dest, pattern, byte_count=None, wait_for=None)
+            queue, dest, pattern, byte_count=byte_count, wait_for=wait_for)
 
 
 def enqueue_svm_migratemem(queue, svms, flags, wait_for=None):
     """
     :arg svms: a collection of Python buffer objects (e.g. :mod:`numpy`
-        arrays), optionally wrapped in :class:`SVM` objects.
+        arrays), or any implementation of :class:`SVMPointer`.
     :arg flags: a combination of :class:`mem_migration_flags`
 
     |std-enqueue-blurb|
@@ -1975,15 +2218,10 @@ def enqueue_svm_migratemem(queue, svms, flags, wait_for=None):
     This function requires OpenCL 2.1.
     """
 
-    return _cl._enqueue_svm_migratemem(
-            queue,
-            [svm.mem if isinstance(svm, SVM) else svm
-                for svm in svms],
-            flags,
-            wait_for)
+    return _cl._enqueue_svm_migratemem(queue, svms, flags, wait_for)
 
 
-def svm_empty(ctx, flags, shape, dtype, order="C", alignment=None):
+def svm_empty(ctx, flags, shape, dtype, order="C", alignment=None, queue=None):
     """Allocate an empty :class:`numpy.ndarray` of the given *shape*, *dtype*
     and *order*. (See :func:`numpy.empty` for the meaning of these arguments.)
     The array will be allocated in shared virtual memory belonging
@@ -2001,6 +2239,10 @@ def svm_empty(ctx, flags, shape, dtype, order="C", alignment=None):
     will likely want to wrap the returned array in an :class:`SVM` tag.
 
     .. versionadded:: 2016.2
+
+    .. versionchanged:: 2022.2
+
+        *queue* argument added.
     """
 
     dtype = np.dtype(dtype)
@@ -2047,7 +2289,9 @@ def svm_empty(ctx, flags, shape, dtype, order="C", alignment=None):
     if alignment is None:
         alignment = itemsize
 
-    svm_alloc = SVMAllocation(ctx, nbytes, alignment, flags, _interface=interface)
+    svm_alloc = _OverriddenArrayInterfaceSVMAllocation(
+            ctx, nbytes, alignment, flags, _interface=interface,
+            queue=queue)
     return np.asarray(svm_alloc)
 
 
@@ -2127,7 +2371,7 @@ def fsvm_empty_like(ctx, ary, alignment=None):
 # }}}
 
 
-_KERNEL_ARG_CLASSES = (
+_KERNEL_ARG_CLASSES: Tuple[type, ...] = (
         MemoryObjectHolder,
         Sampler,
         CommandQueue,
