@@ -23,11 +23,13 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 """
 
-import pyopencl._cl as _cl
 import re
-import sys
 import os
-from pytools import Record
+import sys
+from dataclasses import dataclass
+from typing import List, Optional, Tuple
+
+import pyopencl._cl as _cl
 
 import logging
 logger = logging.getLogger(__name__)
@@ -120,7 +122,6 @@ class CacheLockManager(CleanupBase):
             cleanup_m.register(self)
 
     def clean_up(self):
-        import os
         os.close(self.fd)
         os.unlink(self.lock_file)
 
@@ -145,7 +146,6 @@ class ModuleCacheDirManager(CleanupBase):
         return join(self.path, n)
 
     def reset(self):
-        import os
         _erase_dir(self.path)
         os.mkdir(self.path)
 
@@ -330,8 +330,10 @@ def retrieve_from_cache(cache_dir, cache_key):
 
 # {{{ top-level driver
 
-class _SourceInfo(Record):
-    pass
+@dataclass(frozen=True)
+class _SourceInfo:
+    dependencies: List[Tuple[str, ...]]
+    log: Optional[str]
 
 
 def _create_built_program_from_source_cached(ctx, src, options_bytes,
