@@ -3,9 +3,10 @@ import numpy
 import theano
 
 import pygpu_ndarray as gpu_ndarray
-from gen_elemwise import MyGpuNdArray, elemwise_collapses
-from test_gpu_ndarray import (dtypes_all, enable_double,
+from .gen_elemwise import MyGpuNdArray, elemwise_collapses
+from .test_gpu_ndarray import (dtypes_all, enable_double,
                               gen_gpu_nd_array, product)
+from functools import reduce
 
 
 def rand(shape, dtype):
@@ -324,10 +325,8 @@ def test_sum():
         # 3d, take 2 minutes on a GTX 470
         (0, 0, 0), (0, 1, 0), (0, 0, 1),
         (5, 4, 3), (5, 4, 3), (5, 4, 3),
-        (4095, 2, 33), (2, 4095, 33), (33, 2, 4095),
         (4096, 2, 33), (2, 4096, 33), (33, 2, 4096),
         (4097, 2, 33), (2, 4097, 33), (33, 2, 4097),
-        (4095, 33, 2), (33, 4095, 2), (2, 33, 4095),
         (4096, 33, 2), (33, 4096, 2), (2, 33, 4096),
         (4097, 33, 2), (33, 4097, 2), (2, 33, 4097),
 
@@ -350,7 +349,7 @@ def test_sum():
             *([dtypes] +
               [[False, True]] +
               [[False, True]] +
-              [[False, True]] +
+              [[-1, 2, -2, 1]] +
               [['f', 'c']])):
 
             cpu_val, gpu_val = gen_gpu_nd_array(shape, dtype, off_o,
