@@ -27,20 +27,32 @@ OTHER DEALINGS IN THE SOFTWARE.
 """
 
 
-from pyopencl.tools import context_dependent_memoize
+import enum
+from typing import Any, List, Optional, Tuple, Union
+
 import numpy as np
+
 import pyopencl as cl
-from pytools import memoize_method
-from pyopencl.tools import (dtype_to_ctype, VectorArg, ScalarArg,
+from pyopencl.tools import context_dependent_memoize
+from pyopencl.tools import (
+        dtype_to_ctype, DtypedArgument, VectorArg, ScalarArg,
         KernelTemplateBase, dtype_to_c_struct)
 
+from pytools import memoize_method
+
 
 # {{{ elementwise kernel code generator
 
-def get_elwise_program(context, arguments, operation,
-        name="elwise_kernel", options=None,
-        preamble="", loop_prep="", after_loop="",
-        use_range=False):
+def get_elwise_program(
+        context: cl.Context,
+        arguments: List[DtypedArgument],
+        operation: str, *,
+        name: str = "elwise_kernel",
+        options: Any = None,
+        preamble: str = "",
+        loop_prep: str = "",
+        after_loop: str = "",
+        use_range: bool = False) -> cl.Program:
 
     if use_range:
         body = r"""//CL//
@@ -103,13 +115,18 @@ def get_elwise_program(context, arguments, operation,
             body=body % dict(operation=operation),
             ))
 
-    from pyopencl import Program
-    return Program(context, source).build(options)
+    return cl.Program(context, source).build(options)
 
 
-def get_elwise_kernel_and_types(context, arguments, operation,
-        name="elwise_kernel", options=None, preamble="", use_range=False,
-        **kwargs):
+def get_elwise_kernel_and_types(
+        context: cl.Context,
+        arguments: Union[str, List[DtypedArgument]],
+        operation: str, *,
+        name: str = "elwise_kernel",
+        options: Any = None,
+        preamble: str = "",
+        use_range: bool = False,
+        **kwargs: Any) -> Tuple[cl.Kernel, List[DtypedArgument]]:
 
     from pyopencl.tools import parse_arg_list, get_arg_offset_adjuster_code
     parsed_args = parse_arg_list(arguments, with_offset=True)
@@ -164,8 +181,12 @@ def get_elwise_kernel_and_types(context, arguments, operation,
     return kernel, parsed_args
 
 
-def get_elwise_kernel(context, arguments, operation,
-        name="elwise_kernel", options=None, **kwargs):
+def get_elwise_kernel(
+        context: cl.Context,
+        arguments: Union[str, List[DtypedArgument]],
+        operation: str, *,
+        name: str = "elwise_kernel",
+        options: Any = None, **kwargs: Any) -> cl.Kernel:
     """Return a L{pyopencl.Kernel} that performs the same scalar operation
     on one or several vectors.
     """
@@ -195,16 +216,24 @@ class ElementwiseKernel:
     :arg preamble: a piece of C source code that gets inserted outside of the
         function context in the elementwise operation's kernel source code.
 
-    .. warning :: Using a `return` statement in *operation* will lead to
+    .. warning :: Using a ``return`` statement in *operation* will lead to
         incorrect results, as some elements may never get processed. Use
         ``PYOPENCL_ELWISE_CONTINUE`` instead.
 
     .. versionchanged:: 2013.1
+
         Added ``PYOPENCL_ELWISE_CONTINUE``.
+
+    .. automethod:: __call__
     """
 
-    def __init__(self, context, arguments, operation,
-            name="elwise_kernel", options=None, **kwargs):
+    def __init__(
+            self,
+            context: cl.Context,
+            arguments: Union[str, List[DtypedArgument]],
+            operation: str,
+            name: str = "elwise_kernel",
+            options: Any = None, **kwargs: Any) -> None:
         self.context = context
         self.arguments = arguments
         self.operation = operation
@@ -213,7 +242,7 @@ class ElementwiseKernel:
         self.kwargs = kwargs
 
     @memoize_method
-    def get_kernel(self, use_range):
+    def get_kernel(self, use_range: bool):
         knl, arg_descrs = get_elwise_kernel_and_types(
             self.context, self.arguments, self.operation,
             name=self.name, options=self.options,
@@ -222,33 +251,41 @@ class ElementwiseKernel:
         for arg in arg_descrs:
             if isinstance(arg, VectorArg) and not arg.with_offset:
                 from warnings import warn
-                warn("ElementwiseKernel '%s' used with VectorArgs that do not "
-                        "have offset support enabled. This usage is deprecated. "
-                        "Just pass with_offset=True to VectorArg, everything should "
-                        "sort itself out automatically." % self.name,
+                warn(
+                        f"ElementwiseKernel '{self.name}' used with VectorArgs "
+                        "that do not have offset support enabled. This usage is "
+                        "deprecated. Just pass with_offset=True to VectorArg, "
+                        "everything should sort itself out automatically.",
                         DeprecationWarning)
 
-        if not [i for i, arg in enumerate(arg_descrs)
-                if isinstance(arg, VectorArg)]:
+        if not any(isinstance(arg, VectorArg) for arg in arg_descrs):
             raise RuntimeError(
-                "ElementwiseKernel can only be used with "
-                "functions that have at least one "
-                "vector argument")
+                "ElementwiseKernel can only be used with functions that have "
+                "at least one vector argument")
+
         return knl, arg_descrs
 
-    def __call__(self, *args, **kwargs):
-        repr_vec = None
+    def __call__(self, *args, **kwargs) -> cl.Event:
+        """
+        Invoke the generated scalar kernel.
 
+        The arguments may either be scalars or :class:`pyopencl.array.Array`
+        instances.
+
+        |std-enqueue-blurb|
+        """
         range_ = kwargs.pop("range", None)
         slice_ = kwargs.pop("slice", None)
         capture_as = kwargs.pop("capture_as", None)
+        queue = kwargs.pop("queue", None)
+        wait_for = kwargs.pop("wait_for", None)
+
+        if kwargs:
+            raise TypeError(f"unknown keyword arguments: '{', '.join(kwargs)}'")
 
         use_range = range_ is not None or slice_ is not None
         kernel, arg_descrs = self.get_kernel(use_range)
 
-        queue = kwargs.pop("queue", None)
-        wait_for = kwargs.pop("wait_for", None)
-
         if wait_for is None:
             wait_for = []
         else:
@@ -257,6 +294,7 @@ class ElementwiseKernel:
 
         # {{{ assemble arg array
 
+        repr_vec = None
         invocation_args = []
         for arg, arg_descr in zip(args, arg_descrs):
             if isinstance(arg_descr, VectorArg):
@@ -267,19 +305,17 @@ class ElementwiseKernel:
             else:
                 invocation_args.append(arg)
 
-        # }}}
+        assert repr_vec is not None
 
-        if kwargs:
-            raise TypeError("unknown keyword arguments: '%s'"
-                    % ", ".join(kwargs))
+        # }}}
 
         if queue is None:
             queue = repr_vec.queue
 
         if slice_ is not None:
             if range_ is not None:
-                raise TypeError("may not specify both range and slice "
-                        "keyword arguments")
+                raise TypeError(
+                    "may not specify both range and slice keyword arguments")
 
             range_ = slice(*slice_.indices(repr_vec.size))
 
@@ -322,12 +358,14 @@ class ElementwiseKernel:
 # {{{ template
 
 class ElementwiseTemplate(KernelTemplateBase):
-    def __init__(self,
-            arguments, operation, name="elwise", preamble="",
-            template_processor=None):
-
-        KernelTemplateBase.__init__(self,
-                template_processor=template_processor)
+    def __init__(
+            self,
+            arguments: Union[str, List[DtypedArgument]],
+            operation: str,
+            name: str = "elwise",
+            preamble: str = "",
+            template_processor: Optional[str] = None) -> None:
+        super().__init__(template_processor=template_processor)
         self.arguments = arguments
         self.operation = operation
         self.name = name
@@ -356,25 +394,55 @@ class ElementwiseTemplate(KernelTemplateBase):
 # }}}
 
 
+# {{{ argument kinds
+
+class ArgumentKind(enum.Enum):
+    ARRAY = enum.auto()
+    DEV_SCALAR = enum.auto()
+    SCALAR = enum.auto()
+
+
+def get_argument_kind(v: Any) -> ArgumentKind:
+    from pyopencl.array import Array
+    if isinstance(v, Array):
+        if v.shape == ():
+            return ArgumentKind.DEV_SCALAR
+        else:
+            return ArgumentKind.ARRAY
+    else:
+        return ArgumentKind.SCALAR
+
+
+def get_decl_and_access_for_kind(name: str, kind: ArgumentKind) -> Tuple[str, str]:
+    if kind == ArgumentKind.ARRAY:
+        return f"*{name}", f"{name}[i]"
+    elif kind == ArgumentKind.SCALAR:
+        return f"{name}", name
+    elif kind == ArgumentKind.DEV_SCALAR:
+        return f"*{name}", f"{name}[0]"
+    else:
+        raise AssertionError()
+
+# }}}
+
+
 # {{{ kernels supporting array functionality
 
 @context_dependent_memoize
 def get_take_kernel(context, dtype, idx_dtype, vec_count=1):
-    ctx = {
-            "idx_tp": dtype_to_ctype(idx_dtype),
-            "tp": dtype_to_ctype(dtype),
-            }
+    idx_tp = dtype_to_ctype(idx_dtype)
 
-    args = ([VectorArg(dtype, "dest" + str(i), with_offset=True)
+    args = ([VectorArg(dtype, f"dest{i}", with_offset=True)
              for i in range(vec_count)]
-            + [VectorArg(dtype, "src" + str(i), with_offset=True)
+            + [VectorArg(dtype, f"src{i}", with_offset=True)
                for i in range(vec_count)]
             + [VectorArg(idx_dtype, "idx", with_offset=True)])
     body = (
-            ("%(idx_tp)s src_idx = idx[i];\n" % ctx)
+            f"{idx_tp} src_idx = idx[i];\n"
             + "\n".join(
-                "dest%d[i] = src%d[src_idx];" % (i, i)
-                for i in range(vec_count)))
+                f"dest{i}[i] = src{i}[src_idx];"
+                for i in range(vec_count))
+            )
 
     return get_elwise_kernel(context, args, body,
             preamble=dtype_to_c_struct(context.devices[0], dtype),
@@ -383,37 +451,31 @@ def get_take_kernel(context, dtype, idx_dtype, vec_count=1):
 
 @context_dependent_memoize
 def get_take_put_kernel(context, dtype, idx_dtype, with_offsets, vec_count=1):
-    ctx = {
-            "idx_tp": dtype_to_ctype(idx_dtype),
-            "tp": dtype_to_ctype(dtype),
-            }
+    idx_tp = dtype_to_ctype(idx_dtype)
 
     args = [
-            VectorArg(dtype, "dest%d" % i)
+            VectorArg(dtype, f"dest{i}")
             for i in range(vec_count)
             ] + [
                 VectorArg(idx_dtype, "gmem_dest_idx", with_offset=True),
                 VectorArg(idx_dtype, "gmem_src_idx", with_offset=True),
             ] + [
-                VectorArg(dtype, "src%d" % i, with_offset=True)
+                VectorArg(dtype, f"src{i}", with_offset=True)
                 for i in range(vec_count)
             ] + [
-                ScalarArg(idx_dtype, "offset%d" % i)
+                ScalarArg(idx_dtype, f"offset{i}")
                 for i in range(vec_count) if with_offsets
             ]
 
     if with_offsets:
         def get_copy_insn(i):
-            return ("dest%d[dest_idx] = "
-                    "src%d[src_idx+offset%d];"
-                    % (i, i, i))
+            return f"dest{i}[dest_idx] = src{i}[src_idx + offset{i}];"
     else:
         def get_copy_insn(i):
-            return ("dest%d[dest_idx] = "
-                    "src%d[src_idx];" % (i, i))
+            return f"dest{i}[dest_idx] = src{i}[src_idx];"
 
-    body = (("%(idx_tp)s src_idx = gmem_src_idx[i];\n"
-                "%(idx_tp)s dest_idx = gmem_dest_idx[i];\n" % ctx)
+    body = ((f"{idx_tp} src_idx = gmem_src_idx[i];\n"
+                f"{idx_tp} dest_idx = gmem_dest_idx[i];\n")
             + "\n".join(get_copy_insn(i) for i in range(vec_count)))
 
     return get_elwise_kernel(context, args, body,
@@ -423,18 +485,15 @@ def get_take_put_kernel(context, dtype, idx_dtype, with_offsets, vec_count=1):
 
 @context_dependent_memoize
 def get_put_kernel(context, dtype, idx_dtype, vec_count=1):
-    ctx = {
-            "idx_tp": dtype_to_ctype(idx_dtype),
-            "tp": dtype_to_ctype(dtype),
-            }
+    idx_tp = dtype_to_ctype(idx_dtype)
 
     args = [
-            VectorArg(dtype, "dest%d" % i, with_offset=True)
+            VectorArg(dtype, f"dest{i}", with_offset=True)
             for i in range(vec_count)
             ] + [
                 VectorArg(idx_dtype, "gmem_dest_idx", with_offset=True),
             ] + [
-                VectorArg(dtype, "src%d" % i, with_offset=True)
+                VectorArg(dtype, f"src{i}", with_offset=True)
                 for i in range(vec_count)
             ] + [
                 VectorArg(np.uint8, "use_fill", with_offset=True)
@@ -443,10 +502,10 @@ def get_put_kernel(context, dtype, idx_dtype, vec_count=1):
             ]
 
     body = (
-            "%(idx_tp)s dest_idx = gmem_dest_idx[i];\n" % ctx
+            f"{idx_tp} dest_idx = gmem_dest_idx[i];\n"
             + "\n".join(
-                    "dest{i}[dest_idx] = (use_fill[{i}] ? src{i}[0] : "
-                    "src{i}[i % val_ary_lengths[{i}]]);".format(i=i)
+                    f"dest{i}[dest_idx] = (use_fill[{i}] ? src{i}[0] : "
+                    f"src{i}[i % val_ary_lengths[{i}]]);"
                     for i in range(vec_count)
                     )
             )
@@ -460,10 +519,12 @@ def get_put_kernel(context, dtype, idx_dtype, vec_count=1):
 def get_copy_kernel(context, dtype_dest, dtype_src):
     src = "src[i]"
     if dtype_dest.kind == "c" != dtype_src.kind:
-        src = "{}_fromreal({})".format(complex_dtype_to_name(dtype_dest), src)
+        name = complex_dtype_to_name(dtype_dest)
+        src = f"{name}_fromreal({src})"
 
     if dtype_dest.kind == "c" and dtype_src != dtype_dest:
-        src = "{}_cast({})".format(complex_dtype_to_name(dtype_dest), src),
+        name = complex_dtype_to_name(dtype_dest)
+        src = f"{name}_cast({src})"
 
     if dtype_dest != dtype_src and (
             dtype_dest.kind == "V" or dtype_src.kind == "V"):
@@ -474,18 +535,18 @@ def get_copy_kernel(context, dtype_dest, dtype_src):
                 tp_dest=dtype_to_ctype(dtype_dest),
                 tp_src=dtype_to_ctype(dtype_src),
                 ),
-            "dest[i] = %s" % src,
+            f"dest[i] = {src}",
             preamble=dtype_to_c_struct(context.devices[0], dtype_dest),
             name="copy")
 
 
-def complex_dtype_to_name(dtype):
+def complex_dtype_to_name(dtype) -> str:
     if dtype == np.complex128:
         return "cdouble"
     elif dtype == np.complex64:
         return "cfloat"
     else:
-        raise RuntimeError("invalid complex type")
+        raise RuntimeError(f"invalid complex type: {dtype}")
 
 
 def real_dtype(dtype):
@@ -532,7 +593,7 @@ def get_axpbyz_kernel(context, dtype_x, dtype_y, dtype_z,
                 tp_y=dtype_to_ctype(dtype_y),
                 tp_z=dtype_to_ctype(dtype_z),
                 ),
-            "z[i] = %s" % result,
+            f"z[i] = {result}",
             name="axpbyz")
 
 
@@ -593,7 +654,7 @@ def get_axpbz_kernel(context, dtype_a, dtype_x, dtype_b, dtype_z):
                 tp_b=dtype_to_ctype(dtype_b),
                 tp_z=dtype_to_ctype(dtype_z),
                 ),
-            "z[i] = " + expr,
+            f"z[i] = {expr}",
             name="axpb")
 
 
@@ -626,7 +687,7 @@ def get_multiply_kernel(context, dtype_x, dtype_y, dtype_z,
                 tp_y=dtype_to_ctype(dtype_y),
                 tp_z=dtype_to_ctype(dtype_z),
                 ),
-            "z[i] = %s" % xy,
+            f"z[i] = {xy}",
             name="multiply")
 
 
@@ -645,6 +706,11 @@ def get_divide_kernel(context, dtype_x, dtype_y, dtype_z,
             x = "{}_cast({})".format(complex_dtype_to_name(dtype_z), x)
         if y_is_complex and dtype_y != dtype_z:
             y = "{}_cast({})".format(complex_dtype_to_name(dtype_z), y)
+    else:
+        if dtype_x != dtype_z:
+            x = f"({dtype_to_ctype(dtype_z)}) ({x})"
+        if dtype_y != dtype_z:
+            y = f"({dtype_to_ctype(dtype_z)}) ({y})"
 
     if x_is_complex and y_is_complex:
         xoy = "{}_divide({}, {})".format(complex_dtype_to_name(dtype_z), x, y)
@@ -664,7 +730,7 @@ def get_divide_kernel(context, dtype_x, dtype_y, dtype_z,
                 tp_y=dtype_to_ctype(dtype_y),
                 tp_z=dtype_to_ctype(dtype_z),
                 ),
-            "z[i] = %s" % xoy,
+            f"z[i] = {xoy}",
             name="divide")
 
 
@@ -699,16 +765,14 @@ def get_rdivide_elwise_kernel(context, dtype_x, dtype_y, dtype_z):
                 tp_y=dtype_to_ctype(dtype_y),
                 tp_z=dtype_to_ctype(dtype_z),
                 ),
-            "z[i] = %s" % yox,
+            f"z[i] = {yox}",
             name="divide_r")
 
 
 @context_dependent_memoize
 def get_fill_kernel(context, dtype):
     return get_elwise_kernel(context,
-            "{tp} *z, {tp} a".format(
-                tp=dtype_to_ctype(dtype),
-                ),
+            "{tp} *z, {tp} a".format(tp=dtype_to_ctype(dtype)),
             "z[i] = a",
             preamble=dtype_to_c_struct(context.devices[0], dtype),
             name="fill")
@@ -717,9 +781,7 @@ def get_fill_kernel(context, dtype):
 @context_dependent_memoize
 def get_reverse_kernel(context, dtype):
     return get_elwise_kernel(context,
-            "{tp} *z, {tp} *y".format(
-                tp=dtype_to_ctype(dtype),
-                ),
+            "{tp} *z, {tp} *y".format(tp=dtype_to_ctype(dtype)),
             "z[i] = y[n-1-i]",
             name="reverse")
 
@@ -731,14 +793,14 @@ def get_arange_kernel(context, dtype):
                 "{root}_add(start, {root}_rmul(i, step))"
                 .format(root=complex_dtype_to_name(dtype)))
     else:
-        expr = "start + ((%s) i)*step" % dtype_to_ctype(dtype)
+        expr = f"start + (({dtype_to_ctype(dtype)}) i) * step"
 
     return get_elwise_kernel(context, [
         VectorArg(dtype, "z", with_offset=True),
         ScalarArg(dtype, "start"),
         ScalarArg(dtype, "step"),
         ],
-        "z[i] = " + expr,
+        f"z[i] = {expr}",
         name="arange")
 
 
@@ -747,17 +809,17 @@ def get_pow_kernel(context, dtype_x, dtype_y, dtype_z,
         is_base_array, is_exp_array):
     if is_base_array:
         x = "x[i]"
-        x_ctype = "%(tp_x)s *x"
+        x_ctype = "{tp_x} *x"
     else:
         x = "x"
-        x_ctype = "%(tp_x)s x"
+        x_ctype = "{tp_x} x"
 
     if is_exp_array:
         y = "y[i]"
-        y_ctype = "%(tp_y)s *y"
+        y_ctype = "{tp_y} *y"
     else:
         y = "y"
-        y_ctype = "%(tp_y)s y"
+        y_ctype = "{tp_y} y"
 
     x_is_complex = dtype_x.kind == "c"
     y_is_complex = dtype_y.kind == "c"
@@ -784,12 +846,12 @@ def get_pow_kernel(context, dtype_x, dtype_y, dtype_z,
         result = f"pow({x}, {y})"
 
     return get_elwise_kernel(context,
-            ("%(tp_z)s *z, " + x_ctype + ", "+y_ctype) % {
-                "tp_x": dtype_to_ctype(dtype_x),
-                "tp_y": dtype_to_ctype(dtype_y),
-                "tp_z": dtype_to_ctype(dtype_z),
-                },
-            "z[i] = %s" % result,
+            ("{tp_z} *z, " + x_ctype + ", " + y_ctype).format(
+                tp_x=dtype_to_ctype(dtype_x),
+                tp_y=dtype_to_ctype(dtype_y),
+                tp_z=dtype_to_ctype(dtype_z),
+                ),
+            f"z[i] = {result}",
             name="pow_method")
 
 
@@ -799,7 +861,7 @@ def get_unop_kernel(context, operator, res_dtype, in_dtype):
         VectorArg(res_dtype, "z", with_offset=True),
         VectorArg(in_dtype, "y", with_offset=True),
         ],
-        "z[i] = %s y[i]" % operator,
+        f"z[i] = {operator} y[i]",
         name="unary_op_kernel")
 
 
@@ -810,7 +872,7 @@ def get_array_scalar_binop_kernel(context, operator, dtype_res, dtype_a, dtype_b
         VectorArg(dtype_a, "a", with_offset=True),
         ScalarArg(dtype_b, "b"),
         ],
-        "out[i] = a[i] %s b" % operator,
+        f"out[i] = a[i] {operator} b",
         name="scalar_binop_kernel")
 
 
@@ -835,7 +897,7 @@ def get_array_scalar_comparison_kernel(context, operator, dtype_a):
         VectorArg(dtype_a, "a", with_offset=True),
         ScalarArg(dtype_a, "b"),
         ],
-        "out[i] = a[i] %s b" % operator,
+        f"out[i] = a[i] {operator} b",
         name="scalar_comparison_kernel")
 
 
@@ -846,7 +908,7 @@ def get_array_comparison_kernel(context, operator, dtype_a, dtype_b):
         VectorArg(dtype_a, "a", with_offset=True),
         VectorArg(dtype_b, "b", with_offset=True),
         ],
-        "out[i] = a[i] %s b[i]" % operator,
+        f"out[i] = a[i] {operator} b[i]",
         name="comparison_kernel")
 
 
@@ -859,26 +921,32 @@ def get_unary_func_kernel(context, func_name, in_dtype, out_dtype=None):
         VectorArg(out_dtype, "z", with_offset=True),
         VectorArg(in_dtype, "y", with_offset=True),
         ],
-        "z[i] = %s(y[i])" % func_name,
-        name="%s_kernel" % func_name)
+        f"z[i] = {func_name}(y[i])",
+        name=f"{func_name}_kernel")
 
 
 @context_dependent_memoize
 def get_binary_func_kernel(context, func_name, x_dtype, y_dtype, out_dtype,
                            preamble="", name=None):
+    if name is None:
+        name = func_name
+
     return get_elwise_kernel(context, [
         VectorArg(out_dtype, "z", with_offset=True),
         VectorArg(x_dtype, "x", with_offset=True),
         VectorArg(y_dtype, "y", with_offset=True),
         ],
-        "z[i] = %s(x[i], y[i])" % func_name,
-        name="%s_kernel" % func_name if name is None else name,
+        f"z[i] = {func_name}(x[i], y[i])",
+        name=f"{name}_kernel",
         preamble=preamble)
 
 
 @context_dependent_memoize
 def get_float_binary_func_kernel(context, func_name, x_dtype, y_dtype,
                                  out_dtype, preamble="", name=None):
+    if name is None:
+        name = func_name
+
     if (np.array(0, x_dtype) * np.array(0, y_dtype)).itemsize > 4:
         arg_type = "double"
         preamble = """
@@ -889,13 +957,14 @@ def get_float_binary_func_kernel(context, func_name, x_dtype, y_dtype,
         """ + preamble
     else:
         arg_type = "float"
+
     return get_elwise_kernel(context, [
         VectorArg(out_dtype, "z", with_offset=True),
         VectorArg(x_dtype, "x", with_offset=True),
         VectorArg(y_dtype, "y", with_offset=True),
         ],
         f"z[i] = {func_name}(({arg_type})x[i], ({arg_type})y[i])",
-        name="%s_kernel" % func_name if name is None else name,
+        name=f"{name}_kernel",
         preamble=preamble)
 
 
@@ -945,6 +1014,32 @@ def get_ldexp_kernel(context, out_dtype=np.float32, sig_dtype=np.float32,
         name="ldexp_kernel")
 
 
+@context_dependent_memoize
+def get_minmaximum_kernel(context, minmax, dtype_z, dtype_x, dtype_y,
+        kind_x: ArgumentKind, kind_y: ArgumentKind):
+    if dtype_z.kind == "f":
+        reduce_func = f"f{minmax}_nanprop"
+    elif dtype_z.kind in "iu":
+        reduce_func = minmax
+    else:
+        raise TypeError("unsupported dtype specified")
+
+    tp_x = dtype_to_ctype(dtype_x)
+    tp_y = dtype_to_ctype(dtype_y)
+    tp_z = dtype_to_ctype(dtype_z)
+    decl_x, acc_x = get_decl_and_access_for_kind("x", kind_x)
+    decl_y, acc_y = get_decl_and_access_for_kind("y", kind_y)
+
+    return get_elwise_kernel(context,
+            f"{tp_z} *z, {tp_x} {decl_x}, {tp_y} {decl_y}",
+            f"z[i] = {reduce_func}({acc_x}, {acc_y})",
+            name=f"{minmax}imum",
+            preamble="""
+                #define fmin_nanprop(a, b) (isnan(a) || isnan(b)) ? a+b : fmin(a, b)
+                #define fmax_nanprop(a, b) (isnan(a) || isnan(b)) ? a+b : fmax(a, b)
+                """)
+
+
 @context_dependent_memoize
 def get_bessel_kernel(context, which_func, out_dtype=np.float64,
                       order_dtype=np.int32, x_dtype=np.float64):
@@ -954,15 +1049,15 @@ def get_bessel_kernel(context, which_func, out_dtype=np.float64,
             ScalarArg(order_dtype, "ord_n"),
             VectorArg(x_dtype, "x", with_offset=True),
             ],
-            "z[i] = bessel_%sn(ord_n, x[i])" % which_func,
-            name="bessel_%sn_kernel" % which_func,
-            preamble="""
+            f"z[i] = bessel_{which_func}n(ord_n, x[i])",
+            name=f"bessel_{which_func}n_kernel",
+            preamble=f"""
             #if __OPENCL_C_VERSION__ < 120
             #pragma OPENCL EXTENSION cl_khr_fp64: enable
             #endif
             #define PYOPENCL_DEFINE_CDOUBLE
-            #include <pyopencl-bessel-%s.cl>
-            """ % which_func)
+            #include <pyopencl-bessel-{which_func}.cl>
+            """)
     else:
         if which_func != "j":
             raise NotImplementedError("complex arguments for Bessel Y")
@@ -1065,4 +1160,4 @@ def get_if_positive_kernel(
 
 # }}}
 
-# vim: fdm=marker:filetype=pyopencl
+# vim: fdm=marker
