@@ -11,7 +11,7 @@ Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
 
-    http://www.apache.org/licenses/LICENSE-2.0
+    https://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
@@ -19,19 +19,28 @@ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 
-Derived from code within the Thrust project, https://github.com/thrust/thrust/
+Derived from code within the Thrust project, https://github.com/NVIDIA/thrust
 """
 
+from abc import ABC, abstractmethod
+from dataclasses import dataclass
+from typing import Any, Dict, List, Optional, Set, Tuple, Union
+
 import numpy as np
 
 import pyopencl as cl
-import pyopencl.array  # noqa
-from pyopencl.tools import (dtype_to_ctype, bitlog2,
-        KernelTemplateBase, _process_code_for_macro,
-        get_arg_list_scalar_arg_dtypes,
+import pyopencl.array
+from pyopencl.tools import (
+        KernelTemplateBase,
+        DtypedArgument,
+        bitlog2,
         context_dependent_memoize,
+        dtype_to_ctype,
+        get_arg_list_scalar_arg_dtypes,
+        get_arg_offset_adjuster_code,
+        _process_code_for_macro,
         _NumpyTypesKeyBuilder,
-        get_arg_offset_adjuster_code)
+        )
 
 import pyopencl._mymako as mako
 from pyopencl._cluda import CLUDA_PREAMBLE
@@ -742,7 +751,7 @@ void ${name_prefix}_final_update(
 
 # {{{ helpers
 
-def _round_down_to_power_of_2(val):
+def _round_down_to_power_of_2(val: int) -> int:
     result = 2**bitlog2(val)
     if result > val:
         result >>= 1
@@ -782,7 +791,7 @@ _IGNORED_WORDS = set("""
 
         typedef for endfor if void while endwhile endfor endif else const printf
         None return bool n char true false ifdef pycl_printf str range assert
-        np iinfo max itemsize __packed__ struct restrict
+        np iinfo max itemsize __packed__ struct restrict ptrdiff_t
 
         set iteritems len setdefault
 
@@ -836,10 +845,11 @@ _IGNORED_WORDS = set("""
         """.split())
 
 
-def _make_template(s):
+def _make_template(s: str):
+    import re
     leftovers = set()
 
-    def replace_id(match):
+    def replace_id(match: "re.Match") -> str:
         # avoid name clashes with user code by adding 'psc_' prefix to
         # identifiers.
 
@@ -847,43 +857,28 @@ def _make_template(s):
         if word in _IGNORED_WORDS:
             return word
         elif word in _PREFIX_WORDS:
-            return "psc_"+word
+            return f"psc_{word}"
         else:
             leftovers.add(word)
             return word
 
-    import re
     s = re.sub(r"\b([a-zA-Z0-9_]+)\b", replace_id, s)
-
     if leftovers:
         from warnings import warn
         warn("leftover words in identifier prefixing: " + " ".join(leftovers))
 
-    return mako.template.Template(s, strict_undefined=True)
+    return mako.template.Template(s, strict_undefined=True)     # type: ignore
 
 
-from pytools import Record, RecordWithoutPickling
+@dataclass(frozen=True)
+class _GeneratedScanKernelInfo:
+    scan_src: str
+    kernel_name: str
+    scalar_arg_dtypes: List[Optional[np.dtype]]
+    wg_size: int
+    k_group_size: int
 
-
-class _GeneratedScanKernelInfo(Record):
-
-    __slots__ = [
-            "scan_src",
-            "kernel_name",
-            "scalar_arg_dtypes",
-            "wg_size",
-            "k_group_size"]
-
-    def __init__(self, scan_src, kernel_name, scalar_arg_dtypes, wg_size,
-            k_group_size):
-        Record.__init__(self,
-                scan_src=scan_src,
-                kernel_name=kernel_name,
-                scalar_arg_dtypes=scalar_arg_dtypes,
-                wg_size=wg_size,
-                k_group_size=k_group_size)
-
-    def build(self, context, options):
+    def build(self, context: cl.Context, options: Any) -> "_BuiltScanKernelInfo":
         program = cl.Program(context, self.scan_src).build(options)
         kernel = getattr(program, self.kernel_name)
         kernel.set_scalar_arg_dtypes(self.scalar_arg_dtypes)
@@ -893,40 +888,33 @@ class _GeneratedScanKernelInfo(Record):
                 k_group_size=self.k_group_size)
 
 
-class _BuiltScanKernelInfo(RecordWithoutPickling):
-
-    __slots__ = ["kernel", "wg_size", "k_group_size"]
-
-    def __init__(self, kernel, wg_size, k_group_size):
-        RecordWithoutPickling.__init__(self,
-                kernel=kernel,
-                wg_size=wg_size,
-                k_group_size=k_group_size)
+@dataclass(frozen=True)
+class _BuiltScanKernelInfo:
+    kernel: cl.Kernel
+    wg_size: int
+    k_group_size: int
 
 
-class _GeneratedFinalUpdateKernelInfo(Record):
+@dataclass(frozen=True)
+class _GeneratedFinalUpdateKernelInfo:
+    source: str
+    kernel_name: str
+    scalar_arg_dtypes: List[Optional[np.dtype]]
+    update_wg_size: int
 
-    def __init__(self, source, kernel_name, scalar_arg_dtypes, update_wg_size):
-        Record.__init__(self,
-                source=source,
-                kernel_name=kernel_name,
-                scalar_arg_dtypes=scalar_arg_dtypes,
-                update_wg_size=update_wg_size)
-
-    def build(self, context, options):
+    def build(self,
+              context: cl.Context,
+              options: Any) -> "_BuiltFinalUpdateKernelInfo":
         program = cl.Program(context, self.source).build(options)
         kernel = getattr(program, self.kernel_name)
         kernel.set_scalar_arg_dtypes(self.scalar_arg_dtypes)
         return _BuiltFinalUpdateKernelInfo(kernel, self.update_wg_size)
 
 
-class _BuiltFinalUpdateKernelInfo(RecordWithoutPickling):
-    __slots__ = ["kernel", "update_wg_size"]
-
-    def __init__(self, kernel, update_wg_size):
-        RecordWithoutPickling.__init__(self,
-                kernel=kernel,
-                update_wg_size=update_wg_size)
+@dataclass(frozen=True)
+class _BuiltFinalUpdateKernelInfo:
+    kernel: cl.Kernel
+    update_wg_size: int
 
 # }}}
 
@@ -935,14 +923,25 @@ class ScanPerformanceWarning(UserWarning):
     pass
 
 
-class _GenericScanKernelBase:
+class GenericScanKernelBase(ABC):
     # {{{ constructor, argument processing
 
-    def __init__(self, ctx, dtype,
-            arguments, input_expr, scan_expr, neutral, output_statement,
-            is_segment_start_expr=None, input_fetch_exprs=None,
-            index_dtype=np.int32,
-            name_prefix="scan", options=None, preamble="", devices=None):
+    def __init__(
+            self,
+            ctx: cl.Context,
+            dtype: Any,
+            arguments: Union[str, List[DtypedArgument]],
+            input_expr: str,
+            scan_expr: str,
+            neutral: Optional[str],
+            output_statement: str,
+            is_segment_start_expr: Optional[str] = None,
+            input_fetch_exprs: Optional[List[Tuple[str, str, int]]] = None,
+            index_dtype: Any = None,
+            name_prefix: str = "scan",
+            options: Any = None,
+            preamble: str = "",
+            devices: Optional[cl.Device] = None) -> None:
         """
         :arg ctx: a :class:`pyopencl.Context` within which the code
             for this scan kernel will be generated.
@@ -954,54 +953,54 @@ class _GenericScanKernelBase:
             specified. All types used here must be known to PyOpenCL.
             (see :func:`pyopencl.tools.get_or_register_dtype`).
         :arg scan_expr: The associative, binary operation carrying out the scan,
-            represented as a C string. Its two arguments are available as `a`
-            and `b` when it is evaluated. `b` is guaranteed to be the
-            'element being updated', and `a` is the increment. Thus,
+            represented as a C string. Its two arguments are available as ``a``
+            and ``b`` when it is evaluated. ``b`` is guaranteed to be the
+            'element being updated', and ``a`` is the increment. Thus,
             if some data is supposed to just propagate along without being
-            modified by the scan, it should live in `b`.
+            modified by the scan, it should live in ``b``.
 
             This expression may call functions given in the *preamble*.
 
-            Another value available to this expression is `across_seg_boundary`,
+            Another value available to this expression is ``across_seg_boundary``,
             a C `bool` indicating whether this scan update is crossing a
-            segment boundary, as defined by `is_segment_start_expr`.
+            segment boundary, as defined by ``is_segment_start_expr``.
             The scan routine does not implement segmentation
-            semantics on its own. It relies on `scan_expr` to do this.
-            This value is available (but always `false`) even for a
+            semantics on its own. It relies on ``scan_expr`` to do this.
+            This value is available (but always ``false``) even for a
             non-segmented scan.
 
             .. note::
 
                 In early pre-releases of the segmented scan,
                 segmentation semantics were implemented *without*
-                relying on `scan_expr`.
+                relying on ``scan_expr``.
 
         :arg input_expr: A C expression, encoded as a string, resulting
             in the values to which the scan is applied. This may be used
             to apply a mapping to values stored in *arguments* before being
             scanned. The result of this expression must match *dtype*.
-            The index intended to be mapped is available as `i` in this
+            The index intended to be mapped is available as ``i`` in this
             expression. This expression may also use the variables defined
             by *input_fetch_expr*.
 
             This expression may also call functions given in the *preamble*.
         :arg output_statement: a C statement that writes
-            the output of the scan. It has access to the scan result as `item`,
-            the preceding scan result item as `prev_item`, and the current index
-            as `i`. `prev_item` in a segmented scan will be the neutral element
+            the output of the scan. It has access to the scan result as ``item``,
+            the preceding scan result item as ``prev_item``, and the current index
+            as ``i``. ``prev_item`` in a segmented scan will be the neutral element
             at a segment boundary, not the immediately preceding item.
 
             Using *prev_item* in output statement has a small run-time cost.
-            `prev_item` enables the construction of an exclusive scan.
+            ``prev_item`` enables the construction of an exclusive scan.
 
             For non-segmented scans, *output_statement* may also reference
-            `last_item`, which evaluates to the scan result of the last
+            ``last_item``, which evaluates to the scan result of the last
             array entry.
         :arg is_segment_start_expr: A C expression, encoded as a string,
-            resulting in a C `bool` value that determines whether a new
+            resulting in a C ``bool`` value that determines whether a new
             scan segments starts at index *i*.  If given, makes the scan a
-            segmented scan. Has access to the current index `i`, the result
-            of *input_expr* as `a`, and in addition may use *arguments* and
+            segmented scan. Has access to the current index ``i``, the result
+            of *input_expr* as ``a``, and in addition may use *arguments* and
             *input_fetch_expr* variables just like *input_expr*.
 
             If it returns true, then previous sums will not spill over into the
@@ -1012,8 +1011,8 @@ class _GenericScanKernelBase:
 
                 ARG_NAME_TYPE NAME = ARG_NAME[i+OFFSET];
 
-            `OFFSET` is allowed to be 0 or -1, and `ARG_NAME_TYPE` is the type
-            of `ARG_NAME`.
+            ``OFFSET`` is allowed to be 0 or -1, and ``ARG_NAME_TYPE`` is the type
+            of ``ARG_NAME``.
         :arg preamble: |preamble|
 
         The first array in the argument list determines the size of the index
@@ -1025,6 +1024,9 @@ class _GenericScanKernelBase:
         being processed in the scan.
         """
 
+        if index_dtype is None:
+            index_dtype = np.dtype(np.int32)
+
         if input_fetch_exprs is None:
             input_fetch_exprs = []
 
@@ -1133,8 +1135,9 @@ class _GenericScanKernelBase:
 
     # }}}
 
-    def finish_setup(self):
-        raise NotImplementedError
+    @abstractmethod
+    def finish_setup(self) -> None:
+        pass
 
 
 generic_scan_kernel_cache = WriteOncePersistentDict(
@@ -1142,7 +1145,7 @@ generic_scan_kernel_cache = WriteOncePersistentDict(
         key_builder=_NumpyTypesKeyBuilder())
 
 
-class GenericScanKernel(_GenericScanKernelBase):
+class GenericScanKernel(GenericScanKernelBase):
     """Generates and executes code that performs prefix sums ("scans") on
     arbitrary types, with many possible tweaks.
 
@@ -1159,29 +1162,31 @@ class GenericScanKernel(_GenericScanKernelBase):
         a = cl.array.arange(queue, 10000, dtype=np.int32)
         knl(a, queue=queue)
 
+    .. automethod:: __init__
+    .. automethod:: __call__
     """
 
-    def finish_setup(self):
+    def finish_setup(self) -> None:
         # Before generating the kernel, see if it's cached.
         from pyopencl.cache import get_device_cache_id
         devices_key = tuple(get_device_cache_id(device)
                 for device in self.devices)
 
         cache_key = (self.kernel_key, devices_key)
-
         from_cache = False
 
-        try:
-            result = generic_scan_kernel_cache[cache_key]
-            from_cache = True
-            logger.debug(
+        if not cl._PYOPENCL_NO_CACHE:
+            try:
+                result = generic_scan_kernel_cache[cache_key]
+                from_cache = True
+                logger.debug(
                     "cache hit for generated scan kernel '%s'" % self.name_prefix)
-            (
-                self.first_level_scan_gen_info,
-                self.second_level_scan_gen_info,
-                self.final_update_gen_info) = result
-        except KeyError:
-            pass
+                (
+                    self.first_level_scan_gen_info,
+                    self.second_level_scan_gen_info,
+                    self.final_update_gen_info) = result
+            except KeyError:
+                pass
 
         if not from_cache:
             logger.debug(
@@ -1207,7 +1212,7 @@ class GenericScanKernel(_GenericScanKernelBase):
                 self.context, self.options)
         del self.final_update_gen_info
 
-    def _finish_setup_impl(self):
+    def _finish_setup_impl(self) -> None:
         # {{{ find usable workgroup/k-group size, build first-level scan
 
         trip_count = 0
@@ -1315,7 +1320,7 @@ class GenericScanKernel(_GenericScanKernelBase):
         second_level_arguments = self.parsed_args + [
                 VectorArg(self.dtype, "interval_sums")]
 
-        second_level_build_kwargs = {}
+        second_level_build_kwargs: Dict[str, Optional[str]] = {}
         if self.is_segmented:
             second_level_arguments.append(
                     VectorArg(self.index_dtype,
@@ -1379,12 +1384,14 @@ class GenericScanKernel(_GenericScanKernelBase):
 
     # {{{ scan kernel build/properties
 
-    def get_local_mem_use(self, k_group_size, wg_size, use_bank_conflict_avoidance):
+    def get_local_mem_use(
+            self, k_group_size: int, wg_size: int,
+            use_bank_conflict_avoidance: bool) -> int:
         arg_dtypes = {}
         for arg in self.parsed_args:
             arg_dtypes[arg.name] = arg.dtype
 
-        fetch_expr_offsets = {}
+        fetch_expr_offsets: Dict[str, Set] = {}
         for _name, arg_name, ife_offset in self.input_fetch_exprs:
             fetch_expr_offsets.setdefault(arg_name, set()).add(ife_offset)
 
@@ -1407,10 +1414,17 @@ class GenericScanKernel(_GenericScanKernelBase):
                     for arg_name, ife_offsets in list(fetch_expr_offsets.items())
                     if -1 in ife_offsets or len(ife_offsets) > 1))
 
-    def generate_scan_kernel(self, max_wg_size, arguments, input_expr,
-            is_segment_start_expr, input_fetch_exprs, is_first_level,
-            store_segment_start_flags, k_group_size,
-            use_bank_conflict_avoidance):
+    def generate_scan_kernel(
+            self,
+            max_wg_size: int,
+            arguments: List[DtypedArgument],
+            input_expr: str,
+            is_segment_start_expr: Optional[str],
+            input_fetch_exprs: List[Tuple[str, str, int]],
+            is_first_level: bool,
+            store_segment_start_flags: bool,
+            k_group_size: int,
+            use_bank_conflict_avoidance: bool) -> _GeneratedScanKernelInfo:
         scalar_arg_dtypes = get_arg_list_scalar_arg_dtypes(arguments)
 
         # Empirically found on Nv hardware: no need to be bigger than this size
@@ -1456,7 +1470,25 @@ class GenericScanKernel(_GenericScanKernelBase):
 
     # }}}
 
-    def __call__(self, *args, **kwargs):
+    def __call__(self, *args: Any, **kwargs: Any) -> cl.Event:
+        """
+        |std-enqueue-blurb|
+
+        .. note::
+
+            The returned :class:`pyopencl.Event` corresponds only to part of the
+            execution of the scan. It is not suitable for profiling.
+
+        :arg queue: queue on which to execute the scan. If not given, the
+            queue of the first :class:`pyopencl.array.Array` in *args* is used
+        :arg allocator: an allocator for the temporary arrays and results. If
+            not given, the allocator of the first :class:`pyopencl.array.Array`
+            in *args* is used.
+        :arg size: specify the length of the scan to be carried out. If not
+            given, this length is inferred from the first argument
+        :arg wait_for: a :class:`list` of events to wait for.
+        """
+
         # {{{ argument processing
 
         allocator = kwargs.get("allocator")
@@ -1470,8 +1502,8 @@ class GenericScanKernel(_GenericScanKernelBase):
             wait_for = list(wait_for)
 
         if len(args) != len(self.parsed_args):
-            raise TypeError("expected %d arguments, got %d" %
-                    (len(self.parsed_args), len(args)))
+            raise TypeError(
+                f"expected {len(self.parsed_args)} arguments, got {len(args)}")
 
         first_array = args[self.first_array_idx]
         allocator = allocator or first_array.allocator
@@ -1520,7 +1552,7 @@ class GenericScanKernel(_GenericScanKernelBase):
 
         if self.store_segment_start_flags:
             segment_start_flags = cl.array.empty(
-                    queue, n, dtype=np.bool8,
+                    queue, n, dtype=np.bool_,
                     allocator=allocator)
 
         # }}}
@@ -1649,8 +1681,17 @@ void ${name_prefix}_debug_scan(
 """
 
 
-class GenericDebugScanKernel(_GenericScanKernelBase):
-    def finish_setup(self):
+class GenericDebugScanKernel(GenericScanKernelBase):
+    """
+    Performs the same function and has the same interface as
+    :class:`GenericScanKernel`, but uses a dead-simple, sequential scan.  Works
+    best on CPU platforms, and helps isolate bugs in scans by removing the
+    potential for issues originating in parallel execution.
+
+    .. automethod:: __call__
+    """
+
+    def finish_setup(self) -> None:
         scan_tpl = _make_template(DEBUG_SCAN_TEMPLATE)
         scan_src = str(scan_tpl.render(
             output_statement=self.output_statement,
@@ -1664,15 +1705,16 @@ class GenericDebugScanKernel(_GenericScanKernelBase):
             **self.code_variables))
 
         scan_prg = cl.Program(self.context, scan_src).build(self.options)
-        self.kernel = getattr(
-                scan_prg, self.name_prefix+"_debug_scan")
+        self.kernel = getattr(scan_prg, f"{self.name_prefix}_debug_scan")
         scalar_arg_dtypes = (
                 [None]
                 + get_arg_list_scalar_arg_dtypes(self.parsed_args)
                 + [self.index_dtype])
         self.kernel.set_scalar_arg_dtypes(scalar_arg_dtypes)
 
-    def __call__(self, *args, **kwargs):
+    def __call__(self, *args: Any, **kwargs: Any) -> cl.Event:
+        """See :meth:`GenericScanKernel.__call__`."""
+
         # {{{ argument processing
 
         allocator = kwargs.get("allocator")
@@ -1687,8 +1729,8 @@ class GenericDebugScanKernel(_GenericScanKernelBase):
             wait_for = list(wait_for)
 
         if len(args) != len(self.parsed_args):
-            raise TypeError("expected %d arguments, got %d" %
-                    (len(self.parsed_args), len(args)))
+            raise TypeError(
+                f"expected {len(self.parsed_args)} arguments, got {len(args)}")
 
         first_array = args[self.first_array_idx]
         allocator = allocator or first_array.allocator
@@ -1782,15 +1824,23 @@ class ExclusiveScanKernel(_LegacyScanKernelBase):
 # {{{ template
 
 class ScanTemplate(KernelTemplateBase):
-    def __init__(self,
-            arguments, input_expr, scan_expr, neutral, output_statement,
-            is_segment_start_expr=None, input_fetch_exprs=None,
-            name_prefix="scan", preamble="", template_processor=None):
+    def __init__(
+            self,
+            arguments: Union[str, List[DtypedArgument]],
+            input_expr: str,
+            scan_expr: str,
+            neutral: Optional[str],
+            output_statement: str,
+            is_segment_start_expr: Optional[str] = None,
+            input_fetch_exprs: Optional[List[Tuple[str, str, int]]] = None,
+            name_prefix: str = "scan",
+            preamble: str = "",
+            template_processor: Any = None) -> None:
+        super().__init__(template_processor=template_processor)
 
         if input_fetch_exprs is None:
             input_fetch_exprs = []
 
-        KernelTemplateBase.__init__(self, template_processor=template_processor)
         self.arguments = arguments
         self.input_expr = input_expr
         self.scan_expr = scan_expr
