@@ -72,6 +72,18 @@ Testing
 
 .. autofunction:: pytest_generate_tests_for_pyopencl
 
+Argument Types
+--------------
+
+.. autoclass:: Argument
+.. autoclass:: DtypedArgument
+
+.. autoclass:: VectorArg
+.. autoclass:: ScalarArg
+.. autoclass:: OtherArg
+
+.. autofunction:: parse_arg_list
+
 Device Characterization
 -----------------------
 
@@ -114,8 +126,9 @@ FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 OTHER DEALINGS IN THE SOFTWARE.
 """
 
-
+from abc import ABC, abstractmethod
 from sys import intern
+from typing import Any, List, Optional, Union
 
 # Do not add a pyopencl import here: This will add an import cycle.
 
@@ -126,7 +139,7 @@ from pytools.persistent_dict import KeyBuilder as KeyBuilderBase
 
 import re
 
-from pyopencl.compyte.dtypes import (  # noqa
+from pyopencl.compyte.dtypes import (  # noqa: F401
         get_or_register_dtype, TypeNameNotKnown,
         register_dtype, dtype_to_ctype)
 
@@ -146,14 +159,14 @@ _register_types()
 
 # {{{ imported names
 
-from pyopencl._cl import (  # noqa
+from pyopencl._cl import (  # noqa: F401
         PooledBuffer, AllocatorBase, DeferredAllocator,
         ImmediateAllocator, MemoryPool,
         )
 
 
 if get_cl_header_version() >= (2, 0):
-    from pyopencl._cl import (  # noqa
+    from pyopencl._cl import (  # noqa: F401
             SVMPool,
             PooledSVM,
             SVMAllocator,
@@ -222,6 +235,7 @@ _MEMPOOL_IFACE_DOCS = """
 
 
 def _monkeypatch_docstrings():
+    from pytools.codegen import remove_common_indentation
 
     PooledBuffer.__doc__ = """
     An object representing a :class:`MemoryPool`-based allocation of
@@ -302,7 +316,7 @@ def _monkeypatch_docstrings():
 
     # {{{ MemoryPool
 
-    MemoryPool.__doc__ = """
+    MemoryPool.__doc__ = remove_common_indentation("""
     A memory pool for OpenCL device memory in :class:`pyopencl.Buffer` form.
     *allocator* must be an instance of one of the above classes, and should be
     an :class:`ImmediateAllocator`.  The memory pool assumes that allocation
@@ -327,7 +341,7 @@ def _monkeypatch_docstrings():
         Synonym for :meth:`allocate` to match :class:`AllocatorBase`.
 
         .. versionadded:: 2011.2
-    """ + _MEMPOOL_IFACE_DOCS
+    """) + _MEMPOOL_IFACE_DOCS
 
     # }}}
 
@@ -336,6 +350,8 @@ _monkeypatch_docstrings()
 
 
 def _monkeypatch_svm_docstrings():
+    from pytools.codegen import remove_common_indentation
+
     # {{{ PooledSVM
 
     PooledSVM.__doc__ = """
@@ -367,7 +383,7 @@ def _monkeypatch_svm_docstrings():
 
     .. automethod:: enqueue_release
 
-        Synonymous to :meth;`release`, for consistency with
+        Synonymous to :meth:`release`, for consistency with
         :class:`~pyopencl.SVMAllocation`. Note that, unlike
         :meth:`pyopencl.SVMAllocation.enqueue_release`, specifying a queue
         or events to be waited for is not supported.
@@ -417,7 +433,7 @@ def _monkeypatch_svm_docstrings():
 
     # {{{ SVMPool
 
-    SVMPool.__doc__ = """
+    SVMPool.__doc__ = remove_common_indentation("""
     A memory pool for OpenCL device memory in :ref:`SVM <svm>` form.
     *allocator* must be an instance of :class:`SVMAllocator`.
 
@@ -427,7 +443,7 @@ def _monkeypatch_svm_docstrings():
     .. automethod:: __call__
 
         Return a :class:`PooledSVM` of the given *size*.
-    """ + _MEMPOOL_IFACE_DOCS
+    """) + _MEMPOOL_IFACE_DOCS
 
     # }}}
 
@@ -686,10 +702,10 @@ def pytest_generate_tests_for_pyopencl(metafunc):
         from pyopencl.tools import pytest_generate_tests_for_pyopencl
                 as pytest_generate_tests
 
-    in your `pytest <http://pytest.org>`_ test scripts allows you to use the
-    arguments *ctx_factory*, *device*, or *platform* in your test functions,
-    and they will automatically be run for each OpenCL device/platform in the
-    system, as appropriate.
+    in your `pytest <https://docs.pytest.org/en/latest/>`__ test scripts allows
+    you to use the arguments *ctx_factory*, *device*, or *platform* in your test
+    functions, and they will automatically be run for each OpenCL device/platform
+    in the system, as appropriate.
 
     The following two environment variabls is also supported to control
     device/platform choice::
@@ -714,27 +730,38 @@ def pytest_generate_tests_for_pyopencl(metafunc):
 
 # {{{ C argument lists
 
-class Argument:
-    pass
+class Argument(ABC):
+    """
+    .. automethod:: declarator
+    """
+
+    @abstractmethod
+    def declarator(self) -> str:
+        pass
 
 
 class DtypedArgument(Argument):
-    def __init__(self, dtype, name):
+    """
+    .. attribute:: name
+    .. attribute:: dtype
+    """
+
+    def __init__(self, dtype: Any, name: str) -> None:
         self.dtype = np.dtype(dtype)
         self.name = name
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "{}({!r}, {})".format(
                 self.__class__.__name__,
                 self.name,
                 self.dtype)
 
-    def __eq__(self, other):
+    def __eq__(self, other: Any) -> bool:
         return (type(self) == type(other)
                 and self.dtype == other.dtype
                 and self.name == other.name)
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         return (
                 hash(type(self))
                 ^ hash(self.dtype)
@@ -742,11 +769,16 @@ class DtypedArgument(Argument):
 
 
 class VectorArg(DtypedArgument):
-    def __init__(self, dtype, name, with_offset=False):
+    """Inherits from :class:`DtypedArgument`.
+
+    .. automethod:: __init__
+    """
+
+    def __init__(self, dtype: Any, name: str, with_offset: bool = False) -> None:
         super().__init__(dtype, name)
         self.with_offset = with_offset
 
-    def declarator(self):
+    def declarator(self) -> str:
         if self.with_offset:
             # Two underscores -> less likelihood of a name clash.
             return "__global {} *{}__base, long {}__offset".format(
@@ -756,40 +788,42 @@ class VectorArg(DtypedArgument):
 
         return result
 
-    def __eq__(self, other):
+    def __eq__(self, other) -> bool:
         return (super().__eq__(other)
                 and self.with_offset == other.with_offset)
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         return super().__hash__() ^ hash(self.with_offset)
 
 
 class ScalarArg(DtypedArgument):
+    """Inherits from :class:`DtypedArgument`."""
+
     def declarator(self):
         return "{} {}".format(dtype_to_ctype(self.dtype), self.name)
 
 
 class OtherArg(Argument):
-    def __init__(self, declarator, name):
+    def __init__(self, declarator: str, name: str) -> None:
         self.decl = declarator
         self.name = name
 
-    def declarator(self):
+    def declarator(self) -> str:
         return self.decl
 
-    def __eq__(self, other):
+    def __eq__(self, other) -> bool:
         return (type(self) == type(other)
                 and self.decl == other.decl
                 and self.name == other.name)
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         return (
                 hash(type(self))
                 ^ hash(self.decl)
                 ^ hash(self.name))
 
 
-def parse_c_arg(c_arg, with_offset=False):
+def parse_c_arg(c_arg: str, with_offset: bool = False) -> DtypedArgument:
     for aspace in ["__local", "__constant"]:
         if aspace in c_arg:
             raise RuntimeError("cannot deal with local or constant "
@@ -807,7 +841,9 @@ def parse_c_arg(c_arg, with_offset=False):
     return parse_c_arg_backend(c_arg, ScalarArg, vec_arg_factory)
 
 
-def parse_arg_list(arguments, with_offset=False):
+def parse_arg_list(
+        arguments: Union[str, List[str], List[DtypedArgument]],
+        with_offset: bool = False) -> List[DtypedArgument]:
     """Parse a list of kernel arguments. *arguments* may be a comma-separate
     list of C declarators in a string, a list of strings representing C
     declarators, or :class:`Argument` objects.
@@ -816,11 +852,12 @@ def parse_arg_list(arguments, with_offset=False):
     if isinstance(arguments, str):
         arguments = arguments.split(",")
 
-    def parse_single_arg(obj):
+    def parse_single_arg(obj: Union[str, DtypedArgument]) -> DtypedArgument:
         if isinstance(obj, str):
             from pyopencl.tools import parse_c_arg
             return parse_c_arg(obj, with_offset=with_offset)
         else:
+            assert isinstance(obj, DtypedArgument)
             return obj
 
     return [parse_single_arg(arg) for arg in arguments]
@@ -840,8 +877,10 @@ def get_arg_list_arg_types(arg_types):
     return tuple(result)
 
 
-def get_arg_list_scalar_arg_dtypes(arg_types):
-    result = []
+def get_arg_list_scalar_arg_dtypes(
+        arg_types: List[DtypedArgument]
+        ) -> List[Optional[np.dtype]]:
+    result: List[Optional[np.dtype]] = []
 
     for arg_type in arg_types:
         if isinstance(arg_type, ScalarArg):
@@ -849,9 +888,9 @@ def get_arg_list_scalar_arg_dtypes(arg_types):
         elif isinstance(arg_type, VectorArg):
             result.append(None)
             if arg_type.with_offset:
-                result.append(np.int64)
+                result.append(np.dtype(np.int64))
         else:
-            raise RuntimeError("arg type not understood: %s" % type(arg_type))
+            raise RuntimeError(f"arg type not understood: {type(arg_type)}")
 
     return result
 
@@ -919,7 +958,7 @@ class _CDeclList:
     def add_dtype(self, dtype):
         dtype = np.dtype(dtype)
 
-        if dtype in [np.float64 or np.complex128]:
+        if dtype in (np.float64, np.complex128):
             self.saw_double = True
 
         if dtype.kind == "c":
@@ -952,7 +991,7 @@ class _CDeclList:
     def visit_arguments(self, arguments):
         for arg in arguments:
             dtype = arg.dtype
-            if dtype in [np.float64 or np.complex128]:
+            if dtype in (np.float64, np.complex128):
                 self.saw_double = True
 
             if dtype.kind == "c":
@@ -981,8 +1020,8 @@ class _CDeclList:
 
 @memoize
 def match_dtype_to_c_struct(device, name, dtype, context=None):
-    """Return a tuple `(dtype, c_decl)` such that the C struct declaration
-    in `c_decl` and the structure :class:`numpy.dtype` instance `dtype`
+    """Return a tuple ``(dtype, c_decl)`` such that the C struct declaration
+    in ``c_decl`` and the structure :class:`numpy.dtype` instance ``dtype``
     have the same memory layout.
 
     Note that *dtype* may be modified from the value that was passed in,
@@ -1013,7 +1052,7 @@ def match_dtype_to_c_struct(device, name, dtype, context=None):
         >>> cl.tools.get_or_register_dtype('id_val', dtype)
 
     As this example shows, it is important to call
-    :func:`get_or_register_dtype` on the modified `dtype` returned by this
+    :func:`get_or_register_dtype` on the modified ``dtype`` returned by this
     function, not the original one.
     """
 
@@ -1087,7 +1126,7 @@ def match_dtype_to_c_struct(device, name, dtype, context=None):
     prg = cl.Program(context, src)
     knl = prg.build(devices=[device]).get_size_and_offsets
 
-    import pyopencl.array  # noqa
+    import pyopencl.array  # noqa: F401
     result_buf = cl.array.empty(queue, 1+len(fields), np.uint32)
     knl(queue, (1,), (1,), result_buf.data)
     queue.finish()
