@@ -24,6 +24,7 @@
 // OTHER DEALINGS IN THE SOFTWARE.
 
 
+#include <memory>
 #define NO_IMPORT_ARRAY
 #define PY_ARRAY_UNIQUE_SYMBOL pyopencl_ARRAY_API
 
@@ -64,6 +65,22 @@ namespace pyopencl {
   }
 
 #endif
+
+#if PYOPENCL_CL_VERSION >= 0x2000
+  class svm_pointer_as_buffer
+  {
+    private:
+      svm_pointer &m_ptr;
+
+    public:
+      svm_pointer_as_buffer(svm_pointer &ptr)
+        : m_ptr(ptr)
+      { }
+
+      svm_pointer &ptr() const
+      { return m_ptr; }
+  };
+#endif
 }
 
 
@@ -74,7 +91,7 @@ using namespace pyopencl;
 
 
 
-void pyopencl_expose_part_2(py::module &m)
+void pyopencl_expose_part_2(py::module_ &m)
 {
   // {{{ image
 
@@ -258,7 +275,7 @@ void pyopencl_expose_part_2(py::module &m)
     py::class_<cls>(m, "MemoryMap", py::dynamic_attr())
       .def("release", &cls::release,
           py::arg("queue").none(true)=nullptr,
-          py::arg("wait_for")=py::none()
+          py::arg("wait_for").none(true)=py::none()
           )
       ;
   }
@@ -273,8 +290,8 @@ void pyopencl_expose_part_2(py::module &m)
       py::arg("shape"),
       py::arg("dtype"),
       py::arg("order")="C",
-      py::arg("strides")=py::none(),
-      py::arg("wait_for")=py::none(),
+      py::arg("strides").none(true)=py::none(),
+      py::arg("wait_for").none(true)=py::none(),
       py::arg("is_blocking")=true);
   m.def("enqueue_map_image", enqueue_map_image,
       py::arg("queue"),
@@ -285,52 +302,145 @@ void pyopencl_expose_part_2(py::module &m)
       py::arg("shape"),
       py::arg("dtype"),
       py::arg("order")="C",
-      py::arg("strides")=py::none(),
-      py::arg("wait_for")=py::none(),
+      py::arg("strides").none(true)=py::none(),
+      py::arg("wait_for").none(true)=py::none(),
       py::arg("is_blocking")=true);
 #endif
 
   // }}}
 
-  // {{{ svm
+  // {{{ svm_pointer
 
 #if PYOPENCL_CL_VERSION >= 0x2000
+  {
+    typedef svm_pointer cls;
+    py::class_<cls>(m, "SVMPointer", py::dynamic_attr())
+      // For consistency, it may seem appropriate to use int_ptr here, but
+      // that would work on both buffers and SVM, and passing a buffer pointer to
+      // a kernel is going to lead to a bad time.
+      .def_property_readonly("svm_ptr",
+          [](cls &self) { return (intptr_t) self.svm_ptr(); })
+      .def_property_readonly("size", [](cls &self) -> py::object
+          {
+            try
+            {
+              return py::cast(self.size());
+            }
+            catch (size_not_available)
+            {
+              return py::none();
+            }
+          })
+      .def_property_readonly("buf", [](cls &self) -> svm_pointer_as_buffer * {
+            return new svm_pointer_as_buffer(self);
+          }, py::return_value_policy::reference_internal)
+      ;
+  }
+
+  {
+    typedef svm_pointer_as_buffer cls;
+    py::class_<cls>(m, "_SVMPointerAsBuffer", pybind11::buffer_protocol())
+      .def_buffer([](cls &self) -> pybind11::buffer_info
+          {
+            size_t size;
+            try
+            {
+              size = self.ptr().size();
+            }
+            catch (size_not_available)
+            {
+              throw pyopencl::error("SVMPointer buffer protocol", CL_INVALID_VALUE,
+                  "size of SVM is not known");
+            }
+            return pybind11::buffer_info(
+                // Pointer to buffer
+                self.ptr().svm_ptr(),
+                // Size of one scalar
+                sizeof(unsigned char),
+                // Python struct-style format descriptor
+                pybind11::format_descriptor<unsigned char>::format(),
+                // Number of dimensions
+                1,
+                // Buffer dimensions
+                { size },
+                // Strides (in bytes) for each index
+                { sizeof(unsigned char) }
+                );
+          })
+    ;
+  }
+
+  // }}}
+
+  // {{{ svm_arg_wrapper
+
   {
     typedef svm_arg_wrapper cls;
-    py::class_<cls>(m, "SVM", py::dynamic_attr())
+    py::class_<cls, svm_pointer>(m, "SVM", py::dynamic_attr())
       .def(py::init<py::object>())
       ;
   }
 
+  // }}}
+
+  // {{{ svm_allocation
+
   {
     typedef svm_allocation cls;
-    py::class_<cls>(m, "SVMAllocation", py::dynamic_attr())
-      .def(py::init<std::shared_ptr<context>, size_t, cl_uint, cl_svm_mem_flags>())
+    py::class_<cls, svm_pointer>(m, "SVMAllocation", py::dynamic_attr())
+      .def(py::init<std::shared_ptr<context>, size_t, cl_uint, cl_svm_mem_flags, const command_queue *>(),
+          py::arg("context"),
+          py::arg("size"),
+          py::arg("alignment"),
+          py::arg("flags"),
+          py::arg("queue").none(true)=py::none()
+          )
       .DEF_SIMPLE_METHOD(release)
       .def("enqueue_release", &cls::enqueue_release,
           ":returns: a :class:`pyopencl.Event`\n\n"
-          "|std-enqueue-blurb|")
-      .def("_ptr_as_int", &cls::ptr_as_int)
+          "|std-enqueue-blurb|",
+          py::arg("queue").none(true)=py::none(),
+          py::arg("wait_for").none(true)=py::none()
+          )
       .def(py::self == py::self)
       .def(py::self != py::self)
-      .def("__hash__", &cls::ptr_as_int)
+      .def("__hash__", [](cls &self) { return (intptr_t) self.svm_ptr(); })
+      .def("bind_to_queue", &cls::bind_to_queue,
+          py::arg("queue"))
+      .DEF_SIMPLE_METHOD(unbind_from_queue)
+
+      // only for diagnostic/debugging/testing purposes!
+      .def_property_readonly("_queue",
+          [](cls const &self) -> py::object
+          {
+            cl_command_queue queue = self.queue();
+            if (queue)
+              return py::cast(new command_queue(queue, true));
+            else
+              return py::none();
+          })
       ;
   }
 
+  // }}}
+
+  // {{{ svm operations
+
   m.def("_enqueue_svm_memcpy", enqueue_svm_memcpy,
       py::arg("queue"),
       py::arg("is_blocking"),
       py::arg("dst"),
       py::arg("src"),
-      py::arg("wait_for")=py::none()
+      py::arg("wait_for").none(true)=py::none(),
+      py::arg("byte_count").none(true)=py::none()
       );
 
   m.def("_enqueue_svm_memfill", enqueue_svm_memfill,
       py::arg("queue"),
       py::arg("dst"),
       py::arg("pattern"),
-      py::arg("byte_count")=py::none(),
-      py::arg("wait_for")=py::none()
+      py::arg("byte_count").none(true)=py::none(),
+      py::arg("wait_for").none(true)=py::none()
       );
 
   m.def("_enqueue_svm_map", enqueue_svm_map,
@@ -338,13 +448,14 @@ void pyopencl_expose_part_2(py::module &m)
       py::arg("is_blocking"),
       py::arg("flags"),
       py::arg("svm"),
-      py::arg("wait_for")=py::none()
+      py::arg("wait_for").none(true)=py::none(),
+      py::arg("size").none(true)=py::none()
       );
 
   m.def("_enqueue_svm_unmap", enqueue_svm_unmap,
       py::arg("queue"),
       py::arg("svm"),
-      py::arg("wait_for")=py::none()
+      py::arg("wait_for").none(true)=py::none()
       );
 #endif
 
@@ -352,8 +463,8 @@ void pyopencl_expose_part_2(py::module &m)
   m.def("_enqueue_svm_migrate_mem", enqueue_svm_migratemem,
       py::arg("queue"),
       py::arg("svms"),
-      py::arg("flags")=py::none(),
-      py::arg("wait_for")=py::none()
+      py::arg("flags").none(true)=py::none(),
+      py::arg("wait_for").none(true)=py::none()
       );
 #endif
 
@@ -418,17 +529,17 @@ void pyopencl_expose_part_2(py::module &m)
       .DEF_SIMPLE_METHOD(get_build_info)
       .def("_build", &cls::build,
           py::arg("options")="",
-          py::arg("devices")=py::none())
+          py::arg("devices").none(true)=py::none())
 #if PYOPENCL_CL_VERSION >= 0x1020
       .def("compile", &cls::compile,
           py::arg("options")="",
-          py::arg("devices")=py::none(),
+          py::arg("devices").none(true)=py::none(),
           py::arg("headers")=py::list())
       .def_static("link", &link_program,
           py::arg("context"),
           py::arg("programs"),
           py::arg("options")="",
-          py::arg("devices")=py::none()
+          py::arg("devices").none(true)=py::none()
           )
 #endif
 #if PYOPENCL_CL_VERSION >= 0x2020
@@ -504,7 +615,7 @@ void pyopencl_expose_part_2(py::module &m)
       .def("get_sub_group_info", &cls::get_sub_group_info,
           py::arg("device"),
           py::arg("param"),
-          py::arg("input_value")=py::none()
+          py::arg("input_value").none(true)=py::none()
           )
 #endif
       ;
@@ -520,14 +631,13 @@ void pyopencl_expose_part_2(py::module &m)
       ;
   }
 
-
   m.def("enqueue_nd_range_kernel", enqueue_nd_range_kernel,
       py::arg("queue"),
       py::arg("kernel"),
       py::arg("global_work_size"),
-      py::arg("local_work_size"),
-      py::arg("global_work_offset")=py::none(),
-      py::arg("wait_for")=py::none(),
+      py::arg("local_work_size").none(true),
+      py::arg("global_work_offset").none(true)=py::none(),
+      py::arg("wait_for").none(true)=py::none(),
       py::arg("g_times_l")=false,
       py::arg("allow_empty_ndrange")=false
       );
@@ -600,19 +710,19 @@ void pyopencl_expose_part_2(py::module &m)
   m.def("enqueue_acquire_gl_objects", enqueue_acquire_gl_objects,
       py::arg("queue"),
       py::arg("mem_objects"),
-      py::arg("wait_for")=py::none()
+      py::arg("wait_for").none(true)=py::none()
       );
   m.def("enqueue_release_gl_objects", enqueue_release_gl_objects,
       py::arg("queue"),
       py::arg("mem_objects"),
-      py::arg("wait_for")=py::none()
+      py::arg("wait_for").none(true)=py::none()
       );
 
 #if defined(cl_khr_gl_sharing) && (cl_khr_gl_sharing >= 1)
   m.def("get_gl_context_info_khr", get_gl_context_info_khr,
       py::arg("properties"),
       py::arg("param_name"),
-      py::arg("platform")=py::none()
+      py::arg("platform").none(true)=py::none()
       );
 #endif
 
