@@ -22,12 +22,10 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 """
 
-# avoid spurious: pytest.mark.parametrize is not callable
-# pylint: disable=not-callable
-
 import numpy as np
 import numpy.linalg as la
 import sys
+from itertools import product
 
 import pytest
 
@@ -35,12 +33,13 @@ import pyopencl as cl
 import pyopencl.array as cl_array
 import pyopencl.cltypes as cltypes
 import pyopencl.tools as cl_tools
-from pyopencl.tools import (  # noqa
+from pyopencl.tools import (  # noqa: F401
         pytest_generate_tests_for_pyopencl as pytest_generate_tests)
 from pyopencl.characterize import has_double_support, has_struct_arg_count_bug
 
 from pyopencl.clrandom import RanluxGenerator, PhiloxGenerator, ThreefryGenerator
 import operator
+import platform
 
 _PYPY = cl._PYPY
 
@@ -81,6 +80,8 @@ def make_random_array(queue, dtype, size):
 
 # {{{ dtype-related
 
+# {{{ test_basic_complex
+
 def test_basic_complex(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -91,11 +92,16 @@ def test_basic_complex(ctx_factory):
 
     ary = (rand(queue, shape=(size,), dtype=np.float32).astype(np.complex64)
             + rand(queue, shape=(size,), dtype=np.float32).astype(np.complex64) * 1j)
+    assert ary.dtype != np.dtype(np.complex128)
     c = np.complex64(5+7j)
 
     host_ary = ary.get()
     assert la.norm((ary*c).get() - c*host_ary) < 1e-5 * la.norm(host_ary)
 
+# }}}
+
+
+# {{{ test_mix_complex
 
 def test_mix_complex(ctx_factory):
     context = ctx_factory()
@@ -171,6 +177,10 @@ def test_mix_complex(ctx_factory):
 
                     assert correct
 
+# }}}
+
+
+# {{{ test_pow_neg1_vs_inv
 
 def test_pow_neg1_vs_inv(ctx_factory):
     ctx = ctx_factory()
@@ -193,6 +203,10 @@ def test_pow_neg1_vs_inv(ctx_factory):
     assert la.norm(res1-ref, np.inf) / la.norm(ref) < 1e-13
     assert la.norm(res2-ref, np.inf) / la.norm(ref) < 1e-13
 
+# }}}
+
+
+# {{{ test_vector_fill
 
 def test_vector_fill(ctx_factory):
     context = ctx_factory()
@@ -205,12 +219,21 @@ def test_vector_fill(ctx_factory):
 
     a_gpu = cl_array.zeros(queue, 100, dtype=cltypes.float4)
 
+# }}}
+
+
+# {{{ test_zeros_large_array
 
 def test_zeros_large_array(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
     dev = queue.device
 
+    if dev.platform.vendor == "Intel(R) Corporation" \
+            and platform.system() == "Windows":
+        pytest.xfail("large array fail with out-of-host memory with"
+                "Intel CPU runtime as of 2022-10-05")
+
     size = 2**28 + 1
     if dev.address_bits == 64 and dev.max_mem_alloc_size >= 8 * size:
         # this shouldn't hang/cause errors
@@ -222,6 +245,10 @@ def test_zeros_large_array(ctx_factory):
     else:
         pass
 
+# }}}
+
+
+# {{{ test_absrealimag
 
 def test_absrealimag(ctx_factory):
     context = ctx_factory()
@@ -252,6 +279,10 @@ def test_absrealimag(ctx_factory):
                 print(dev_res-host_res)
             assert correct
 
+# }}}
+
+
+# {{{ test_custom_type_zeros
 
 def test_custom_type_zeros(ctx_factory):
     context = ctx_factory()
@@ -281,6 +312,10 @@ def test_custom_type_zeros(ctx_factory):
 
     assert np.array_equal(np.zeros(n, dtype), z)
 
+# }}}
+
+
+# {{{ test_custom_type_fill
 
 def test_custom_type_fill(ctx_factory):
     context = ctx_factory()
@@ -310,6 +345,10 @@ def test_custom_type_fill(ctx_factory):
 
     assert np.array_equal(np.zeros(n, dtype), z)
 
+# }}}
+
+
+# {{{ test_custom_type_take_put
 
 def test_custom_type_take_put(ctx_factory):
     context = ctx_factory()
@@ -341,9 +380,13 @@ def test_custom_type_take_put(ctx_factory):
 
 # }}}
 
+# }}}
+
 
 # {{{ operators
 
+# {{{ test_div_type_matches_numpy
+
 @pytest.mark.parametrize("dtype", [np.int8, np.int32, np.int64, np.float32])
 # FIXME Implement florodiv
 #@pytest.mark.parametrize("op", [operator.truediv, operator.floordiv])
@@ -359,6 +402,10 @@ def test_div_type_matches_numpy(ctx_factory, dtype, op):
     assert res_np.dtype == res.dtype
     assert np.allclose(res_np, res.get())
 
+# }}}
+
+
+# {{{ test_rmul_yields_right_type
 
 def test_rmul_yields_right_type(ctx_factory):
     context = ctx_factory()
@@ -373,6 +420,10 @@ def test_rmul_yields_right_type(ctx_factory):
     two_a = np.float32(2)*a_gpu
     assert isinstance(two_a, cl_array.Array)
 
+# }}}
+
+
+# {{{ test_pow_array
 
 def test_pow_array(ctx_factory):
     context = ctx_factory()
@@ -387,6 +438,10 @@ def test_pow_array(ctx_factory):
     result = (a_gpu ** a_gpu).get()
     assert (np.abs(pow(a, a) - result) < 3e-3).all()
 
+# }}}
+
+
+# {{{ test_pow_number
 
 def test_pow_number(ctx_factory):
     context = ctx_factory()
@@ -398,6 +453,10 @@ def test_pow_number(ctx_factory):
     result = pow(a_gpu, 2).get()
     assert (np.abs(a ** 2 - result) < 1e-3).all()
 
+# }}}
+
+
+# {{{ test_multiply
 
 def test_multiply(ctx_factory):
     """Test the muliplication of an array with a scalar. """
@@ -417,6 +476,10 @@ def test_multiply(ctx_factory):
 
                 assert (a * scalar == a_mult).all()
 
+# }}}
+
+
+# {{{ test_multiply_array
 
 def test_multiply_array(ctx_factory):
     """Test the multiplication of two arrays."""
@@ -434,6 +497,11 @@ def test_multiply_array(ctx_factory):
     assert (a * a == a_squared).all()
 
 
+# }}}
+
+
+# {{{ test_addition_array
+
 def test_addition_array(ctx_factory):
     """Test the addition of two arrays."""
 
@@ -446,6 +514,10 @@ def test_addition_array(ctx_factory):
 
     assert (a + a == a_added).all()
 
+# }}}
+
+
+# {{{ test_addition_scalar
 
 def test_addition_scalar(ctx_factory):
     """Test the addition of an array and a scalar."""
@@ -459,6 +531,10 @@ def test_addition_scalar(ctx_factory):
 
     assert (7 + a == a_added).all()
 
+# }}}
+
+
+# {{{ test_subtract_array
 
 @pytest.mark.parametrize(("dtype_a", "dtype_b"),
         [
@@ -469,7 +545,7 @@ def test_addition_scalar(ctx_factory):
             (np.int64, np.uint32),
             ])
 def test_subtract_array(ctx_factory, dtype_a, dtype_b):
-    """Test the substraction of two arrays."""
+    """Test the subtraction of two arrays."""
     #test data
     a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(dtype_a)
     b = np.array([10, 20, 30, 40, 50,
@@ -487,9 +563,13 @@ def test_subtract_array(ctx_factory, dtype_a, dtype_b):
     result = (b_gpu - a_gpu).get()
     assert (b - a == result).all()
 
+# }}}
 
-def test_substract_scalar(ctx_factory):
-    """Test the substraction of an array and a scalar."""
+
+# {{{ test_subtract_scalar
+
+def test_subtract_scalar(ctx_factory):
+    """Test the subtraction of an array and a scalar."""
 
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -506,6 +586,10 @@ def test_substract_scalar(ctx_factory):
     result = (7 - a_gpu).get()
     assert (7 - a == result).all()
 
+# }}}
+
+
+# {{{ test_divide_scalar
 
 def test_divide_scalar(ctx_factory):
     """Test the division of an array and a scalar."""
@@ -540,6 +624,10 @@ def test_divide_scalar(ctx_factory):
         assert (np.abs(c_gpu.get() - c) < 1e-3).all()
         assert c_gpu.dtype is c.dtype
 
+# }}}
+
+
+# {{{ test_divide_array
 
 def test_divide_array(ctx_factory):
     """Test the division of an array and a scalar. """
@@ -571,6 +659,10 @@ def test_divide_array(ctx_factory):
         assert (np.abs(d_gpu.get() - d) < 1e-3).all()
         assert d_gpu.dtype is d.dtype
 
+# }}}
+
+
+# {{{ test_divide_inplace_scalar
 
 def test_divide_inplace_scalar(ctx_factory):
     """Test inplace division of arrays and a scalar."""
@@ -607,6 +699,10 @@ def test_divide_inplace_scalar(ctx_factory):
             assert (np.abs(a_gpu.get() - a) < 1e-3).all()
             assert a_gpu.dtype is a.dtype
 
+# }}}
+
+
+# {{{ test_divide_inplace_array
 
 def test_divide_inplace_array(ctx_factory):
     """Test inplace division of arrays."""
@@ -645,6 +741,10 @@ def test_divide_inplace_array(ctx_factory):
             assert (np.abs(a_gpu.get() - a) < 1e-3).all()
             assert a_gpu.dtype is a.dtype
 
+# }}}
+
+
+# {{{ test_bitwise
 
 def test_bitwise(ctx_factory):
     if _PYPY:
@@ -662,8 +762,6 @@ def test_bitwise(ctx_factory):
     for a_dtype, b_dtype in product(dtypes, dtypes):
         ary_len = 16
 
-        np.random.seed(10)
-
         int32_min = np.iinfo(np.int32).min
         int32_max = np.iinfo(np.int32).max
 
@@ -722,9 +820,13 @@ def test_bitwise(ctx_factory):
 
 # }}}
 
+# }}}
+
 
 # {{{ RNG
 
+# {{{ test_random_float_in_range
+
 @pytest.mark.parametrize("rng_class",
         [RanluxGenerator, PhiloxGenerator, ThreefryGenerator])
 @pytest.mark.parametrize("ary_size", [300, 301, 302, 303, 10007, 1000000])
@@ -788,6 +890,10 @@ def test_random_float_in_range(ctx_factory, rng_class, ary_size, plot_hist=False
             pt.hist(ran.get(), 30)
             pt.show()
 
+# }}}
+
+
+# {{{ test_random_int_in_range
 
 @pytest.mark.parametrize("dtype", [np.int32, np.int64])
 @pytest.mark.parametrize("rng_class",
@@ -825,9 +931,13 @@ def test_random_int_in_range(ctx_factory, rng_class, dtype, plot_hist=False):
 
 # }}}
 
+# }}}
+
 
 # {{{ misc
 
+# {{{ test_numpy_integer_shape
+
 def test_numpy_integer_shape(ctx_factory):
     try:
         list(np.int32(17))
@@ -842,6 +952,37 @@ def test_numpy_integer_shape(ctx_factory):
     cl_array.empty(queue, np.int32(17), np.float32)
     cl_array.empty(queue, (np.int32(17), np.int32(17)), np.float32)
 
+# }}}
+
+
+# {{{ test_allocation_with_various_shape_scalar_types
+
+def test_allocation_with_various_shape_scalar_types(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    dims_ok = (2, np.int32(7), np.uint64(1))
+    dims_not_ok = (-1, 5.70, np.float32(7))
+
+    shapes_ok_1d = list(product(dims_ok))
+    shapes_ok_2d = list(product(dims_ok, dims_ok))
+    shapes_ok_3d = list(product(dims_ok, dims_ok, dims_ok))
+
+    shapes_not_ok_1d = list(product(dims_not_ok))
+    shapes_not_ok_2d = list(product(dims_ok, dims_not_ok))
+    shapes_not_ok_3d = list(product(dims_not_ok, dims_not_ok, dims_not_ok))
+
+    for shape in shapes_ok_1d + shapes_ok_2d + shapes_ok_3d:
+        cl_array.empty(queue, shape, np.float32)
+
+    for shape in shapes_not_ok_1d + shapes_not_ok_2d + shapes_not_ok_3d:
+        with pytest.raises(ValueError):
+            cl_array.empty(queue, shape, np.float32)
+
+# }}}
+
+
+# {{{ test_len
 
 def test_len(ctx_factory):
     context = ctx_factory()
@@ -851,6 +992,10 @@ def test_len(ctx_factory):
     a_cpu = cl_array.to_device(queue, a)
     assert len(a_cpu) == 10
 
+# }}}
+
+
+# {{{ test_stride_preservation
 
 def test_stride_preservation(ctx_factory):
     if _PYPY:
@@ -859,21 +1004,28 @@ def test_stride_preservation(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
 
-    a = np.random.rand(3, 3)
+    rng = np.random.default_rng(seed=42)
+    a = rng.random(size=(3, 3))
+
     at = a.T
     print(at.flags.f_contiguous, at.flags.c_contiguous)
     at_gpu = cl_array.to_device(queue, at)
     print(at_gpu.flags.f_contiguous, at_gpu.flags.c_contiguous)
     assert np.allclose(at_gpu.get(), at)
 
+# }}}
+
+
+# {{{ test_nan_arithmetic
 
 def test_nan_arithmetic(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
+    rng = np.random.default_rng(seed=42)
 
     def make_nan_contaminated_vector(size):
-        shape = (size,)
-        a = np.random.randn(*shape).astype(np.float32)
+        a = rng.standard_normal(size=(size,), dtype=np.float32)
+
         from random import randrange
         for _i in range(size // 10):
             a[randrange(0, size)] = float("nan")
@@ -891,6 +1043,10 @@ def test_nan_arithmetic(ctx_factory):
 
     assert (np.isnan(ab) == np.isnan(ab_gpu)).all()
 
+# }}}
+
+
+# {{{ test_mem_pool_with_arrays
 
 def test_mem_pool_with_arrays(ctx_factory):
     context = ctx_factory()
@@ -903,6 +1059,10 @@ def test_mem_pool_with_arrays(ctx_factory):
     assert a_dev.allocator is mem_pool
     assert b_dev.allocator is mem_pool
 
+# }}}
+
+
+# {{{ test_view
 
 def test_view(ctx_factory):
     context = ctx_factory()
@@ -923,6 +1083,10 @@ def test_view(ctx_factory):
     view = a_dev.view(np.int16)
     assert view.shape == (8, 32) and view.dtype == np.int16
 
+# }}}
+
+
+# {{{ test_diff
 
 def test_diff(ctx_factory):
     context = ctx_factory()
@@ -938,6 +1102,10 @@ def test_diff(ctx_factory):
             cl.array.diff(a_dev).get() - np.diff(a))
     assert err < 1e-4
 
+# }}}
+
+
+# {{{ test_copy
 
 def test_copy(ctx_factory):
     context = ctx_factory()
@@ -966,9 +1134,13 @@ def test_copy(ctx_factory):
 
 # }}}
 
+# }}}
+
 
 # {{{ slices, concatenation
 
+# {{{ test_slice
+
 def test_slice(ctx_factory):
     if _PYPY:
         pytest.xfail("numpypy: spurious as_strided failure")
@@ -986,10 +1158,17 @@ def test_slice(ctx_factory):
     a = a_gpu.get()
     b = b_gpu.get()
 
+    if queue.device.platform.name == "Intel(R) OpenCL":
+        pytest.skip("Intel CL regularly crashes on this test case "
+            "-- https://github.com/conda-forge/"
+            "intel-compiler-repack-feedstock/issues/7")
+    else:
+        start_offset = 0
+
     from random import randrange
     for _i in range(20):
-        start = randrange(ary_len)
-        end = randrange(start, ary_len)
+        start = randrange(ary_len - start_offset)
+        end = randrange(start+start_offset, ary_len)
 
         a_gpu_slice = tp(2)*a_gpu[start:end]
         a_slice = tp(2)*a[start:end]
@@ -997,8 +1176,9 @@ def test_slice(ctx_factory):
         assert la.norm(a_gpu_slice.get() - a_slice) == 0
 
     for _i in range(20):
-        start = randrange(ary_len)
-        end = randrange(start, ary_len)
+        start = randrange(ary_len-start_offset)
+        #end = randrange(start+start_offset, ary_len)
+        end = start
 
         a_gpu[start:end] = tp(2)*b[start:end]
         a[start:end] = tp(2)*b[start:end]
@@ -1006,14 +1186,18 @@ def test_slice(ctx_factory):
         assert la.norm(a_gpu.get() - a) == 0
 
     for _i in range(20):
-        start = randrange(ary_len)
-        end = randrange(start, ary_len)
+        start = randrange(ary_len-start_offset)
+        end = randrange(start+start_offset, ary_len)
 
         a_gpu[start:end] = tp(2)*b_gpu[start:end]
         a[start:end] = tp(2)*b[start:end]
 
         assert la.norm(a_gpu.get() - a) == 0
 
+# }}}
+
+
+# {{{ test_concatenate
 
 def test_concatenate(ctx_factory):
     context = ctx_factory()
@@ -1035,9 +1219,13 @@ def test_concatenate(ctx_factory):
 
 # }}}
 
+# }}}
+
 
 # {{{ conditionals, any, all
 
+# {{{ test_comparisons
+
 def test_comparisons(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -1073,6 +1261,11 @@ def test_comparisons(ctx_factory):
         assert (res2_dev.get() == res2).all()
 
 
+# }}}
+
+
+# {{{ test_any_all
+
 def test_any_all(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -1095,6 +1288,10 @@ def test_any_all(ctx_factory):
 
 # }}}
 
+# }}}
+
+
+# {{{ test_map_to_host
 
 def test_map_to_host(ctx_factory):
     if _PYPY:
@@ -1125,11 +1322,14 @@ def test_map_to_host(ctx_factory):
 
     assert (a_host_saved == a_dev.get()).all()
 
+# }}}
+
+
+# {{{ test_view_and_strides
 
 def test_view_and_strides(ctx_factory):
     if _PYPY:
         pytest.xfail("numpypy: no array creation from __array_interface__")
-    return
 
     context = ctx_factory()
     queue = cl.CommandQueue(context)
@@ -1146,6 +1346,10 @@ def test_view_and_strides(ctx_factory):
     with pytest.raises(AssertionError):
         assert (yv.get() == x.get()[:3, :5]).all()
 
+# }}}
+
+
+# {{{ test_meshmode_view
 
 def test_meshmode_view(ctx_factory):
     if _PYPY:
@@ -1168,6 +1372,10 @@ def test_meshmode_view(ctx_factory):
     x = result.get()
     assert (view(x) == 1).all()
 
+# }}}
+
+
+# {{{ test_event_management
 
 def test_event_management(ctx_factory):
     context = ctx_factory()
@@ -1207,6 +1415,10 @@ def test_event_management(ctx_factory):
 
     assert len(x.events) < 100
 
+# }}}
+
+
+# {{{ test_reshape
 
 def test_reshape(ctx_factory):
     context = ctx_factory()
@@ -1229,6 +1441,10 @@ def test_reshape(ctx_factory):
     with pytest.raises(ValueError):
         a_dev.reshape(-1, -1, 4)
 
+# }}}
+
+
+# {{{ test_skip_slicing
 
 def test_skip_slicing(ctx_factory):
     context = ctx_factory()
@@ -1240,8 +1456,13 @@ def test_skip_slicing(ctx_factory):
     a = cl_array.to_device(queue, a_host)
     b = a[::3]
     assert b.shape == b_host.shape
-    assert np.array_equal(b[1].get(), b_host[1])  # noqa pylint:disable=unsubscriptable-object
+    # pylint:disable=unsubscriptable-object
+    assert np.array_equal(b[1].get(), b_host[1])
 
+# }}}
+
+
+# {{{ test_transpose
 
 def test_transpose(ctx_factory):
     if _PYPY:
@@ -1259,6 +1480,10 @@ def test_transpose(ctx_factory):
     #assert np.allclose(a_gpu.transpose((1,2,0)).get(), a.transpose((1,2,0)))
     assert np.array_equal(a_gpu.T.get(), a.T)
 
+# }}}
+
+
+# {{{ test_newaxis
 
 def test_newaxis(ctx_factory):
     context = ctx_factory()
@@ -1277,13 +1502,18 @@ def test_newaxis(ctx_factory):
         if b.shape[i] > 1:
             assert b_gpu.strides[i] == b.strides[i]
 
+# }}}
+
+
+# {{{ test_squeeze
 
 def test_squeeze(ctx_factory):
     context = ctx_factory()
     queue = cl.CommandQueue(context)
+    rng = np.random.default_rng(seed=42)
 
     shape = (40, 2, 5, 100)
-    a_cpu = np.random.random(size=shape)
+    a_cpu = rng.random(size=shape)
     a_gpu = cl_array.to_device(queue, a_cpu)
 
     # Slice with length 1 on dimensions 0 and 1
@@ -1312,6 +1542,10 @@ def test_squeeze(ctx_factory):
     # Check that we get the original values out
     #assert np.all(a_gpu_slice.get().ravel() == a_gpu_squeezed_slice.get().ravel())
 
+# }}}
+
+
+# {{{ test_fancy_fill
 
 def test_fancy_fill(ctx_factory):
     if _PYPY:
@@ -1331,17 +1565,22 @@ def test_fancy_fill(ctx_factory):
 
     assert np.all(numpy_dest == cl_dest.get())
 
+# }}}
+
+
+# {{{ test_fancy_indexing
 
 def test_fancy_indexing(ctx_factory):
     if _PYPY:
         pytest.xfail("numpypy: multi value setting is not supported")
     context = ctx_factory()
     queue = cl.CommandQueue(context)
+    rng = np.random.default_rng(seed=42)
 
     n = 2 ** 20 + 2**18 + 22
     numpy_dest = np.zeros(n, dtype=np.int32)
     numpy_idx = np.arange(n, dtype=np.int32)
-    np.random.shuffle(numpy_idx)
+    rng.shuffle(numpy_idx)
     numpy_src = 20000+np.arange(n, dtype=np.int32)
 
     cl_dest = cl_array.to_device(queue, numpy_dest)
@@ -1358,6 +1597,10 @@ def test_fancy_indexing(ctx_factory):
 
     assert np.array_equal(numpy_dest, cl_dest.get())
 
+# }}}
+
+
+# {{{ test_multi_put
 
 def test_multi_put(ctx_factory):
     if _PYPY:
@@ -1384,6 +1627,10 @@ def test_multi_put(ctx_factory):
 
     assert np.all(np.all(out_compare[i] == out_arrays[i].get()) for i in range(9))
 
+# }}}
+
+
+# {{{ test_get_async
 
 def test_get_async(ctx_factory):
     context = ctx_factory()
@@ -1395,7 +1642,8 @@ def test_get_async(ctx_factory):
         pytest.xfail("the async get test fails on POCL + Nvidia,"
                 "at least the K40, as of pocl 1.6, 2021-01-20")
 
-    a = np.random.rand(10**6).astype(np.dtype("float32"))
+    rng = np.random.default_rng(seed=42)
+    a = rng.random(10**6, dtype=np.float32)
     a_gpu = cl_array.to_device(queue, a)
     b = a + a**5 + 1
     b_gpu = a_gpu + a_gpu**5 + 1
@@ -1416,6 +1664,10 @@ def test_get_async(ctx_factory):
     evt.wait()
     assert np.abs(b1 - b).mean() < 1e-5
 
+# }}}
+
+
+# {{{ test_outoforderqueue_get
 
 def test_outoforderqueue_get(ctx_factory):
     context = ctx_factory()
@@ -1424,13 +1676,19 @@ def test_outoforderqueue_get(ctx_factory):
                properties=cl.command_queue_properties.OUT_OF_ORDER_EXEC_MODE_ENABLE)
     except Exception:
         pytest.skip("out-of-order queue not available")
-    a = np.random.rand(10**6).astype(np.dtype("float32"))
+
+    rng = np.random.default_rng(seed=42)
+    a = rng.random(10**6, dtype=np.float32)
     a_gpu = cl_array.to_device(queue, a)
     b_gpu = a_gpu + a_gpu**5 + 1
     b1 = b_gpu.get()  # testing that this waits for events
     b = a + a**5 + 1
     assert np.abs(b1 - b).mean() < 1e-5
 
+# }}}
+
+
+# {{{ test_outoforderqueue_copy
 
 def test_outoforderqueue_copy(ctx_factory):
     context = ctx_factory()
@@ -1439,7 +1697,9 @@ def test_outoforderqueue_copy(ctx_factory):
                properties=cl.command_queue_properties.OUT_OF_ORDER_EXEC_MODE_ENABLE)
     except Exception:
         pytest.skip("out-of-order queue not available")
-    a = np.random.rand(10**6).astype(np.dtype("float32"))
+
+    rng = np.random.default_rng(seed=42)
+    a = rng.random(10**6, dtype=np.float32)
     a_gpu = cl_array.to_device(queue, a)
     c_gpu = a_gpu**2 - 7
     b_gpu = c_gpu.copy()  # testing that this waits for and creates events
@@ -1450,6 +1710,11 @@ def test_outoforderqueue_copy(ctx_factory):
     assert np.abs(b1 - b).mean() < 1e-5
 
 
+# }}}
+
+
+# {{{ test_outoforderqueue_indexing
+
 def test_outoforderqueue_indexing(ctx_factory):
     context = ctx_factory()
     try:
@@ -1457,8 +1722,11 @@ def test_outoforderqueue_indexing(ctx_factory):
                properties=cl.command_queue_properties.OUT_OF_ORDER_EXEC_MODE_ENABLE)
     except Exception:
         pytest.skip("out-of-order queue not available")
-    a = np.random.rand(10**6).astype(np.dtype("float32"))
-    i = (8e5 + 1e5 * np.random.rand(10**5)).astype(np.dtype("int32"))
+
+    rng = np.random.default_rng(seed=42)
+    a = rng.random(10**6, dtype=np.float32)
+    i = (8e5 + 1e5 * rng.random(10**5)).astype(np.int32)
+
     a_gpu = cl_array.to_device(queue, a)
     i_gpu = cl_array.to_device(queue, i)
     c_gpu = (a_gpu**2)[i_gpu - 10000]
@@ -1472,6 +1740,10 @@ def test_outoforderqueue_indexing(ctx_factory):
     b[i + 10000] = c - 10
     assert np.abs(b1 - b).mean() < 1e-5
 
+# }}}
+
+
+# {{{ test_outoforderqueue_reductions
 
 def test_outoforderqueue_reductions(ctx_factory):
     context = ctx_factory()
@@ -1481,7 +1753,9 @@ def test_outoforderqueue_reductions(ctx_factory):
     except Exception:
         pytest.skip("out-of-order queue not available")
     # 0/1 values to avoid accumulated rounding error
-    a = (np.random.rand(10**6) > 0.5).astype(np.dtype("float32"))
+    rng = np.random.default_rng(seed=42)
+    a = (rng.random(10**6) > 0.5).astype(np.float32)
+
     a[800000] = 10  # all<5 looks true until near the end
     a_gpu = cl_array.to_device(queue, a)
     b1 = cl_array.sum(a_gpu).get()
@@ -1489,6 +1763,10 @@ def test_outoforderqueue_reductions(ctx_factory):
     b3 = (a_gpu < 5).all().get()
     assert b1 == a.sum() and b2 == a.dot(3 - a) and b3 == 0
 
+# }}}
+
+
+# {{{ test_negative_dim_rejection
 
 def test_negative_dim_rejection(ctx_factory):
     context = ctx_factory()
@@ -1508,6 +1786,10 @@ def test_negative_dim_rejection(ctx_factory):
         with pytest.raises(ValueError):
             cl_array.Array(queue, shape=(-1, right_dim), dtype=np.float64)
 
+# }}}
+
+
+# {{{ test_zero_size_array
 
 @pytest.mark.parametrize("empty_shape", [0, (), (3, 0, 2), (0, 5), (5, 0)])
 def test_zero_size_array(ctx_factory, empty_shape):
@@ -1535,6 +1817,10 @@ def test_zero_size_array(ctx_factory, empty_shape):
         assert c_flat.flags.c_contiguous == c_host_flat.flags.c_contiguous
         assert c_flat.flags.f_contiguous == c_host_flat.flags.f_contiguous
 
+# }}}
+
+
+# {{{ test_str_without_queue
 
 def test_str_without_queue(ctx_factory):
     context = ctx_factory()
@@ -1544,6 +1830,10 @@ def test_str_without_queue(ctx_factory):
     print(str(a))
     print(repr(a))
 
+# }}}
+
+
+# {{{ test_stack
 
 @pytest.mark.parametrize("order", ("F", "C"))
 @pytest.mark.parametrize("input_dims", (1, 2, 3))
@@ -1556,8 +1846,7 @@ def test_stack(ctx_factory, input_dims, order):
     shape = (2, 2, 2)[:input_dims]
     axis = -1 if order == "F" else 0
 
-    from numpy.random import default_rng
-    rng = default_rng()
+    rng = np.random.default_rng(seed=42)
     x_in = rng.random(size=shape)
     y_in = rng.random(size=shape)
     x_in = x_in if order == "C" else np.asfortranarray(x_in)
@@ -1569,6 +1858,10 @@ def test_stack(ctx_factory, input_dims, order):
     np.testing.assert_allclose(cla.stack((x, y), axis=axis).get(),
                                 np.stack((x_in, y_in), axis=axis))
 
+# }}}
+
+
+# {{{ test_assign_different_strides
 
 def test_assign_different_strides(ctx_factory):
     cl_ctx = ctx_factory()
@@ -1581,11 +1874,16 @@ def test_assign_different_strides(ctx_factory):
     with pytest.raises(NotImplementedError):
         b[:] = a
 
+# }}}
+
+
+# {{{ test_branch_operations_on_pure_scalars
 
 def test_branch_operations_on_pure_scalars():
-    x = np.random.rand()
-    y = np.random.rand()
-    cond = np.random.choice([False, True])
+    rng = np.random.default_rng(seed=42)
+    x = rng.random()
+    y = rng.random()
+    cond = rng.choice([False, True])
 
     np.testing.assert_allclose(np.maximum(x, y),
                                cl_array.maximum(x, y))
@@ -1594,23 +1892,78 @@ def test_branch_operations_on_pure_scalars():
     np.testing.assert_allclose(np.where(cond, x, y),
                                cl_array.if_positive(cond, x, y))
 
+# }}}
+
+
+# {{{ test_branch_operations_on_nans
+
+@pytest.mark.parametrize("op", [
+    cl_array.maximum,
+    cl_array.minimum,
+])
+@pytest.mark.parametrize("special_a", [
+    np.nan,
+    np.inf,
+    -np.inf,
+])
+@pytest.mark.parametrize("special_b", [
+    np.nan,
+    np.inf,
+    -np.inf,
+    None
+])
+def test_branch_operations_on_nans(ctx_factory, op, special_a, special_b):
+    ctx = ctx_factory()
+    cq = cl.CommandQueue(ctx)
+
+    def sb_or(x):
+        if special_b is None:
+            return x
+        else:
+            return special_b
+
+    x_np = np.array([special_a, sb_or(1.), special_a, sb_or(2.), sb_or(3.)],
+            dtype=np.float64)
+    y_np = np.array([special_a, special_a, sb_or(1.), sb_or(3.), sb_or(2.)],
+            dtype=np.float64)
+
+    x_cl = cl_array.to_device(cq, x_np)
+    y_cl = cl_array.to_device(cq, y_np)
+
+    ref = getattr(np, op.__name__)(x_np, y_np)
+    result = op(x_cl, y_cl)
+    if isinstance(result, cl_array.Array):
+        result = result.get()
+
+    np.testing.assert_allclose(result, ref)
+
+# }}}
+
+
+# {{{ test_slice_copy
 
 def test_slice_copy(ctx_factory):
     cl_ctx = ctx_factory()
     queue = cl.CommandQueue(cl_ctx)
 
-    x = cl.array.to_device(queue, np.random.rand(96, 27))
+    rng = np.random.default_rng(seed=42)
+    x = cl.array.to_device(queue, rng.random(size=(96, 27)))
     y = x[::8, ::3]
     with pytest.raises(RuntimeError):
         y.copy()
 
+# }}}
+
+
+# {{{{ test_ravel
 
 @pytest.mark.parametrize("order", ("C", "F"))
 def test_ravel(ctx_factory, order):
     ctx = ctx_factory()
     cq = cl.CommandQueue(ctx)
 
-    x = np.random.randn(10, 4)
+    rng = np.random.default_rng(seed=42)
+    x = rng.standard_normal(size=(10, 4))
 
     if order == "F":
         x = np.asfortranarray(x)
@@ -1624,6 +1977,10 @@ def test_ravel(ctx_factory, order):
     np.testing.assert_allclose(x_cl.ravel(order=order).get(),
                                x.ravel(order=order))
 
+# }}}
+
+
+# {{{ test_arithmetic_on_non_scalars
 
 def test_arithmetic_on_non_scalars(ctx_factory):
     pytest.importorskip("dataclasses")
@@ -1642,16 +1999,19 @@ def test_arithmetic_on_non_scalars(ctx_factory):
     with pytest.raises(TypeError):
         ArrayContainer(np.ones(100)) + cl.array.zeros(cq, (10,), dtype=np.float64)
 
+# }}}
+
+
+# {{{ test_arithmetic_with_device_scalars
 
 @pytest.mark.parametrize("which", ("add", "sub", "mul", "truediv"))
 def test_arithmetic_with_device_scalars(ctx_factory, which):
     import operator
-    from numpy.random import default_rng
 
     ctx = ctx_factory()
     cq = cl.CommandQueue(ctx)
 
-    rng = default_rng()
+    rng = np.random.default_rng(seed=42)
     ndim = rng.integers(1, 5)
 
     shape = tuple(rng.integers(2, 7) for i in range(ndim))
@@ -1666,6 +2026,10 @@ def test_arithmetic_with_device_scalars(ctx_factory, which):
 
     np.testing.assert_allclose(res_cl.get(), res_np)
 
+# }}}
+
+
+# {{{ test_if_positive_with_scalars
 
 @pytest.mark.parametrize("then_type", ["array", "host_scalar", "device_scalar"])
 @pytest.mark.parametrize("else_type", ["array", "host_scalar", "device_scalar"])
@@ -1673,7 +2037,7 @@ def test_if_positive_with_scalars(ctx_factory, then_type, else_type):
     ctx = ctx_factory()
     cq = cl.CommandQueue(ctx)
 
-    rng = np.random.default_rng()
+    rng = np.random.default_rng(seed=42)
     shape = (512,)
 
     criterion_np = rng.random(shape)
@@ -1701,6 +2065,10 @@ def test_if_positive_with_scalars(ctx_factory, then_type, else_type):
 
     np.testing.assert_allclose(result_cl.get(), result_np)
 
+# }}}
+
+
+# {{{ test_maximum_minimum_with_scalars
 
 def test_maximum_minimum_with_scalars(ctx_factory):
     ctx = ctx_factory()
@@ -1726,6 +2094,27 @@ def test_maximum_minimum_with_scalars(ctx_factory):
     result = cl_array.minimum(a_cl, b_cl, queue=cq)
     np.testing.assert_allclose(result.get(), b_np)
 
+    # Test 'untyped' scalars
+    # FIXME: these don't work with unsized ints
+    result = cl_array.minimum(4.0, b_cl, queue=cq)
+    np.testing.assert_allclose(result.get(), b_np)
+    result = cl_array.maximum(4.0, b_cl, queue=cq)
+    np.testing.assert_allclose(result.get(), a_np)
+
+    result = cl_array.minimum(b_cl, 4.0, queue=cq)
+    np.testing.assert_allclose(result.get(), b_np)
+    result = cl_array.maximum(b_cl, 4.0, queue=cq)
+    np.testing.assert_allclose(result.get(), a_np)
+
+    result = cl_array.minimum(-3.0, 4.0, queue=cq)
+    np.testing.assert_allclose(result, b_np)
+    result = cl_array.maximum(-3.0, 4.0, queue=cq)
+    np.testing.assert_allclose(result, a_np)
+
+# }}}
+
+
+# {{{ test_empty_reductions_vs_numpy
 
 @pytest.mark.parametrize(("reduction", "supports_initial"), [
     (cl_array.any, False),
@@ -1788,6 +2177,124 @@ def test_empty_reductions_vs_numpy(ctx_factory, reduction, supports_initial):
 
     # }}}
 
+# }}}
+
+
+# {{{ test_reduction_nan_handling
+
+@pytest.mark.parametrize("with_initial", [False, True])
+@pytest.mark.parametrize("input_case", ["only nans", "mixed"])
+@pytest.mark.parametrize("reduction", [
+    cl_array.sum,
+    cl_array.max,
+    cl_array.min,
+    ])
+def test_reduction_nan_handling(ctx_factory, reduction, input_case, with_initial):
+    ctx = ctx_factory()
+    cq = cl.CommandQueue(ctx)
+
+    if input_case == "only nans":
+        x_np = np.array([np.nan, np.nan], dtype=np.float64)
+    elif input_case == "mixed":
+        x_np = np.array([np.nan, 1.], dtype=np.float64)
+    else:
+        raise ValueError("invalid input case")
+
+    x_cl = cl_array.to_device(cq, x_np)
+
+    if with_initial:
+        ref = getattr(np, reduction.__name__)(x_np, initial=5.0)
+        result = reduction(x_cl, initial=5.0)
+    else:
+        ref = getattr(np, reduction.__name__)(x_np)
+        result = reduction(x_cl)
+
+    if isinstance(result, cl_array.Array):
+        result = result.get()
+
+    np.testing.assert_allclose(result, ref)
+
+# }}}
+
+
+# {{{ test_reductions_dtype
+
+def test_dtype_conversions(ctx_factory):
+    ctx = ctx_factory()
+    queue = cl.CommandQueue(ctx)
+
+    ary = cl.array.to_device(queue, np.linspace(0, 1, 32))
+
+    for func, nargs, arg_name in [
+            (cl.array.sum, 1, "dtype"),
+            (cl.array.dot, 2, "dtype"),
+            (cl.array.vdot, 2, "dtype"),
+            (cl.array.cumsum, 1, "output_dtype"),
+            ]:
+        for dtype in [np.float32, np.float64]:
+            result = func(*((ary,) * nargs), **{arg_name: dtype})
+            assert result.dtype == dtype, result.dtype
+
+# }}}
+
+
+# {{{ test_svm_mem_pool_with_arrays
+
+@pytest.mark.parametrize("use_mempool", [False, True])
+def test_arrays_with_svm_allocators(ctx_factory, use_mempool):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+    queue2 = cl.CommandQueue(context)
+
+    from pyopencl.characterize import has_coarse_grain_buffer_svm
+    has_cg_svm = has_coarse_grain_buffer_svm(queue.device)
+
+    if not has_cg_svm:
+        pytest.skip("Need coarse-grained SVM support for this test.")
+
+    alloc = cl_tools.SVMAllocator(context, queue=queue)
+    if use_mempool:
+        alloc = cl_tools.SVMPool(alloc)
+
+    def alloc2(size):
+        allocation = alloc(size)
+        allocation.bind_to_queue(queue2)
+        return allocation
+
+    a_dev = cl_array.arange(queue, 2000, dtype=np.float32, allocator=alloc)
+    b_dev = cl_array.to_device(queue, np.arange(2000), allocator=alloc) + 4000
+
+    assert a_dev.allocator is alloc
+    assert b_dev.allocator is alloc
+
+    assert a_dev.data._queue == queue
+    assert b_dev.data._queue == queue
+
+    a_dev2 = cl_array.arange(queue2, 2000, dtype=np.float32, allocator=alloc2)
+    b_dev2 = cl_array.to_device(queue2, np.arange(2000), allocator=alloc2) + 4000
+
+    assert a_dev2.allocator is alloc2
+    assert b_dev2.allocator is alloc2
+
+    assert a_dev2.data._queue == queue2
+    assert b_dev2.data._queue == queue2
+
+    np.testing.assert_allclose((a_dev+b_dev).get(), (a_dev2+b_dev2).get())
+
+    with pytest.warns(cl_array.InconsistentOpenCLQueueWarning):
+        a_dev2.with_queue(queue)
+
+        # safe to let this proceed to deallocation, since we're not
+        # operating on the memory
+
+    with pytest.warns(cl_array.InconsistentOpenCLQueueWarning):
+        cl_array.empty(queue2, 2000, np.float32, allocator=alloc)
+
+        # safe to let this proceed to deallocation, since we're not
+        # operating on the memory
+
+# }}}
+
 
 if __name__ == "__main__":
     if len(sys.argv) > 1:
@@ -1796,4 +2303,4 @@ if __name__ == "__main__":
         from pytest import main
         main([__file__])
 
-# vim: filetype=pyopencl:fdm=marker
+# vim: fdm=marker
