@@ -20,10 +20,6 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 """
 
-# avoid spurious: pytest.mark.parametrize is not callable
-# pylint: disable=not-callable
-
-
 import numpy as np
 import numpy.linalg as la
 import pytest
@@ -32,19 +28,11 @@ import pyopencl as cl
 import pyopencl.array as cl_array
 import pyopencl.cltypes as cltypes
 import pyopencl.clrandom
-from pyopencl.tools import (  # noqa
+from pyopencl.tools import (  # noqa: F401
         pytest_generate_tests_for_pyopencl as pytest_generate_tests,
         ImmediateAllocator, DeferredAllocator)
 from pyopencl.characterize import get_pocl_version
 
-# Are CL implementations crashy? You be the judge. :)
-try:
-    import faulthandler  # noqa
-except ImportError:
-    pass
-else:
-    faulthandler.enable()
-
 
 def _xfail_if_pocl(plat, up_to_version, msg="unsupported by pocl"):
     if plat.vendor == "The pocl project":
@@ -59,6 +47,8 @@ def _xfail_if_pocl_gpu(device, what):
                 "at least the Titan V, as of pocl 1.6, 2021-01-20")
 
 
+# {{{ test_get_info
+
 def test_get_info(ctx_factory):
     ctx = ctx_factory()
     device, = ctx.devices
@@ -79,7 +69,7 @@ def test_get_info(ctx_factory):
             (cl.Program, cl.program_info.KERNEL_NAMES),
             (cl.Program, cl.program_info.NUM_KERNELS),
         ])
-    CRASH_QUIRKS = [  # noqa
+    CRASH_QUIRKS = [  # noqa: N806
             (("NVIDIA Corporation", "NVIDIA CUDA",
                 "OpenCL 1.0 CUDA 3.0.1"),
                 [
@@ -114,7 +104,7 @@ def test_get_info(ctx_factory):
                     (cl.Program, cl.program_info.SOURCE),
                     ]),
             ]
-    QUIRKS = []  # noqa
+    QUIRKS = []  # noqa: N806
 
     def find_quirk(quirk_list, cl_obj, info):
         for (vendor, name, version), quirks in quirk_list:
@@ -236,6 +226,10 @@ def test_get_info(ctx_factory):
         do_test(img, cl.image_info,
                 lambda info: img.get_image_info(info))
 
+# }}}
+
+
+# {{{ test_int_ptr
 
 def test_int_ptr(ctx_factory):
     def do_test(obj):
@@ -285,6 +279,10 @@ def test_int_ptr(ctx_factory):
         img = cl.Image(ctx, cl.mem_flags.READ_ONLY, img_format, (128, 256))
         do_test(img)
 
+# }}}
+
+
+# {{{ test_invalid_kernel_names_cause_failures
 
 def test_invalid_kernel_names_cause_failures(ctx_factory):
     ctx = ctx_factory()
@@ -308,6 +306,10 @@ def test_invalid_kernel_names_cause_failures(ctx_factory):
         else:
             raise
 
+# }}}
+
+
+# {{{ test_image_format_constructor
 
 def test_image_format_constructor():
     # doesn't need image support to succeed
@@ -319,6 +321,10 @@ def test_image_format_constructor():
     if not cl._PYPY:
         assert not hasattr(iform, "__dict__")
 
+# }}}
+
+
+# {{{ test_device_topology_amd_constructor
 
 def test_device_topology_amd_constructor():
     # doesn't need cl_amd_device_attribute_query support to succeed
@@ -331,6 +337,10 @@ def test_device_topology_amd_constructor():
     if not cl._PYPY:
         assert not hasattr(topol, "__dict__")
 
+# }}}
+
+
+# {{{ test_nonempty_supported_image_formats
 
 def test_nonempty_supported_image_formats(ctx_factory):
     context = ctx_factory()
@@ -344,6 +354,10 @@ def test_nonempty_supported_image_formats(ctx_factory):
         from pytest import skip
         skip("images not supported on %s" % device.name)
 
+# }}}
+
+
+# {{{ test_that_python_args_fail
 
 def test_that_python_args_fail(ctx_factory):
     context = ctx_factory()
@@ -353,7 +367,8 @@ def test_that_python_args_fail(ctx_factory):
         { a[get_global_id(0)] *= (b+c); }
         """).build()
 
-    a = np.random.rand(50000)
+    rng = np.random.default_rng(seed=42)
+    a = rng.random(50000)
     queue = cl.CommandQueue(context)
     mf = cl.mem_flags
     a_buf = cl.Buffer(context, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=a)
@@ -378,6 +393,10 @@ def test_that_python_args_fail(ctx_factory):
     a_result = np.empty_like(a)
     cl.enqueue_copy(queue, a_buf, a_result).wait()
 
+# }}}
+
+
+# {{{ test_image_2d
 
 def test_image_2d(ctx_factory):
     context = ctx_factory()
@@ -413,7 +432,9 @@ def test_image_2d(ctx_factory):
         """).build()
 
     num_channels = 1
-    a = np.random.rand(1024, 512, num_channels).astype(np.float32)
+
+    rng = np.random.default_rng(seed=42)
+    a = rng.random((1024, 512, num_channels), dtype=np.float32)
     if num_channels == 1:
         a = a[:, :, 0]
 
@@ -449,6 +470,10 @@ def test_image_2d(ctx_factory):
         else:
             assert good
 
+# }}}
+
+
+# {{{ test_image_3d
 
 def test_image_3d(ctx_factory):
     #test for image_from_array for 3d image of float2
@@ -489,7 +514,9 @@ def test_image_3d(ctx_factory):
 
     num_channels = 2
     shape = (3, 4, 2)
-    a = np.random.random(shape + (num_channels,)).astype(np.float32)
+
+    rng = np.random.default_rng(seed=42)
+    a = rng.random(size=shape + (num_channels,), dtype=np.float32)
 
     queue = cl.CommandQueue(context)
     try:
@@ -525,6 +552,10 @@ def test_image_3d(ctx_factory):
         else:
             assert good
 
+# }}}
+
+
+# {{{ test_copy_buffer
 
 def test_copy_buffer(ctx_factory):
     context = ctx_factory()
@@ -532,7 +563,8 @@ def test_copy_buffer(ctx_factory):
     queue = cl.CommandQueue(context)
     mf = cl.mem_flags
 
-    a = np.random.rand(50000).astype(np.float32)
+    rng = np.random.default_rng(seed=42)
+    a = rng.random(50000, dtype=np.float32)
     b = np.empty_like(a)
 
     buf1 = cl.Buffer(context, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
@@ -543,6 +575,10 @@ def test_copy_buffer(ctx_factory):
 
     assert la.norm(a - b) == 0
 
+# }}}
+
+
+# {{{ test_mempool_*
 
 def test_mempool(ctx_factory):
     from pyopencl.tools import MemoryPool, ImmediateAllocator
@@ -595,6 +631,10 @@ def test_mempool_32bit_issues():
         for offs in range(-5, 5):
             pool.allocate(2**i + offs)
 
+# }}}
+
+
+# {{{ test_allocator
 
 @pytest.mark.parametrize("allocator_cls", [ImmediateAllocator, DeferredAllocator])
 def test_allocator(ctx_factory, allocator_cls):
@@ -612,6 +652,10 @@ def test_allocator(ctx_factory, allocator_cls):
     assert mem is not None
     assert mem2 is None
 
+# }}}
+
+
+# {{{ test_vector_args
 
 def test_vector_args(ctx_factory):
     context = ctx_factory()
@@ -633,7 +677,10 @@ def test_vector_args(ctx_factory):
 
     assert (dest == x).all()
 
+# }}}
 
+
+# {{{ test_header_dep_handling
 def test_header_dep_handling(ctx_factory):
     context = ctx_factory()
 
@@ -651,6 +698,10 @@ def test_header_dep_handling(ctx_factory):
     cl.Program(context, kernel_src).build(["-I", dirname(__file__)])
     cl.Program(context, kernel_src).build(["-I", dirname(__file__)])
 
+# }}}
+
+
+# {{{ test_context_dep_memoize
 
 def test_context_dep_memoize(ctx_factory):
     context = ctx_factory()
@@ -668,6 +719,10 @@ def test_context_dep_memoize(ctx_factory):
 
     assert counter[0] == 1
 
+# }}}
+
+
+# {{{ test_can_build_and_run_binary
 
 def test_can_build_and_run_binary(ctx_factory):
     ctx = ctx_factory()
@@ -692,6 +747,10 @@ def test_can_build_and_run_binary(ctx_factory):
 
     foo.simple(queue, (n,), (16,), a_dev.data, dest_dev.data)
 
+# }}}
+
+
+# {{{ test_enqueue_barrier_marker
 
 def test_enqueue_barrier_marker(ctx_factory):
     ctx = ctx_factory()
@@ -710,6 +769,10 @@ def test_enqueue_barrier_marker(ctx_factory):
     evt2 = cl.enqueue_marker(queue, wait_for=[evt1])
     cl.enqueue_barrier(queue, wait_for=[evt1, evt2])
 
+# }}}
+
+
+# {{{ test_wait_for_events
 
 def test_wait_for_events(ctx_factory):
     ctx = ctx_factory()
@@ -718,6 +781,10 @@ def test_wait_for_events(ctx_factory):
     evt2 = cl.enqueue_marker(queue)
     cl.wait_for_events([evt1, evt2])
 
+# }}}
+
+
+# {{{ test_unload_compiler
 
 def test_unload_compiler(platform):
     if (platform._get_cl_version() < (1, 2)
@@ -730,6 +797,10 @@ def test_unload_compiler(platform):
         skip("Intel proprietary driver does not support unloading compiler")
     cl.unload_platform_compiler(platform)
 
+# }}}
+
+
+# {{{ test_platform_get_devices
 
 def test_platform_get_devices(ctx_factory):
     ctx = ctx_factory()
@@ -756,6 +827,10 @@ def test_platform_get_devices(ctx_factory):
         for dev in devs:
             assert dev.type & dev_type == dev_type
 
+# }}}
+
+
+# {{{ test_user_event
 
 def test_user_event(ctx_factory):
     ctx = ctx_factory()
@@ -802,6 +877,10 @@ def test_user_event(ctx_factory):
         raise RuntimeError("cl.wait_for_events timeout on UserEvent")
     assert evt.command_execution_status == cl.command_execution_status.COMPLETE
 
+# }}}
+
+
+# {{{ test_buffer_get_host_array
 
 def test_buffer_get_host_array(ctx_factory):
     if cl._PYPY:
@@ -811,7 +890,8 @@ def test_buffer_get_host_array(ctx_factory):
     ctx = ctx_factory()
     mf = cl.mem_flags
 
-    host_buf = np.random.rand(25).astype(np.float32)
+    rng = np.random.default_rng(seed=42)
+    host_buf = rng.random(25, dtype=np.float32)
     buf = cl.Buffer(ctx, mf.READ_WRITE | mf.USE_HOST_PTR, hostbuf=host_buf)
     host_buf2 = buf.get_host_array(25, np.float32)
     assert (host_buf == host_buf2).all()
@@ -827,7 +907,7 @@ def test_buffer_get_host_array(ctx_factory):
     except cl.LogicError:
         pass
 
-    host_buf = np.random.rand(25).astype(np.float32)
+    host_buf = rng.random(25, dtype=np.float32)
     buf = cl.Buffer(ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=host_buf)
     try:
         host_buf2 = buf.get_host_array(25, np.float32)
@@ -836,6 +916,10 @@ def test_buffer_get_host_array(ctx_factory):
     except cl.LogicError:
         pass
 
+# }}}
+
+
+# {{{ test_program_valued_get_info
 
 def test_program_valued_get_info(ctx_factory):
     ctx = ctx_factory()
@@ -853,6 +937,10 @@ def test_program_valued_get_info(ctx_factory):
     assert knl.program == prg
     knl.program.binaries[0]
 
+# }}}
+
+
+# {{{ test_event_set_callback
 
 def test_event_set_callback(ctx_factory):
     import sys
@@ -867,8 +955,9 @@ def test_event_set_callback(ctx_factory):
     if ctx._get_cl_version() < (1, 1):
         pytest.skip("OpenCL 1.1 or newer required for set_callback")
 
-    a_np = np.random.rand(50000).astype(np.float32)
-    b_np = np.random.rand(50000).astype(np.float32)
+    rng = np.random.default_rng(seed=42)
+    a_np = rng.random(50000, dtype=np.float32)
+    b_np = rng.random(50000, dtype=np.float32)
 
     got_called = []
 
@@ -913,6 +1002,10 @@ def test_event_set_callback(ctx_factory):
 
     assert got_called
 
+# }}}
+
+
+# {{{ test_global_offset
 
 def test_global_offset(ctx_factory):
     context = ctx_factory()
@@ -926,7 +1019,9 @@ def test_global_offset(ctx_factory):
         """).build()
 
     n = 50
-    a = np.random.rand(n).astype(np.float32)
+
+    rng = np.random.default_rng(seed=42)
+    a = rng.random(n, dtype=np.float32)
 
     queue = cl.CommandQueue(context)
     mf = cl.mem_flags
@@ -941,6 +1036,10 @@ def test_global_offset(ctx_factory):
 
     assert (a_2 == 2*a).all()
 
+# }}}
+
+
+# {{{ test_sub_buffers
 
 def test_sub_buffers(ctx_factory):
     ctx = ctx_factory()
@@ -954,7 +1053,9 @@ def test_sub_buffers(ctx_factory):
     queue = cl.CommandQueue(ctx)
 
     n = 30000
-    a = (np.random.rand(n) * 100).astype(np.uint8)
+
+    rng = np.random.default_rng(seed=42)
+    a = (rng.random(n) * 100).astype(np.uint8)
 
     mf = cl.mem_flags
     a_buf = cl.Buffer(ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=a)
@@ -969,6 +1070,10 @@ def test_sub_buffers(ctx_factory):
 
     assert np.array_equal(a_sub, a_sub_ref)
 
+# }}}
+
+
+# {{{ test_spirv
 
 def test_spirv(ctx_factory):
     ctx = ctx_factory()
@@ -988,7 +1093,11 @@ def test_spirv(ctx_factory):
     b_dev = cl.clrandom.rand(queue, n, np.float32)
     dest_dev = cl_array.empty_like(a_dev)
 
-    with open("add-vectors-%d.spv" % queue.device.address_bits, "rb") as spv_file:
+    from os.path import dirname, join
+    spv_filename = join(dirname(__file__),
+            "add-vectors-%d.spv" % queue.device.address_bits)
+
+    with open(spv_filename, "rb") as spv_file:
         spv = spv_file.read()
 
     prg = cl.Program(ctx, spv).build()
@@ -1000,16 +1109,19 @@ def test_spirv(ctx_factory):
 
     assert la.norm((dest_dev - (a_dev+b_dev)).get()) < 1e-7
 
+# }}}
 
-def test_coarse_grain_svm(ctx_factory):
+
+# {{{ test_coarse_grain_svm
+
+@pytest.mark.parametrize("use_opaque_style", [False, True])
+def test_coarse_grain_svm(ctx_factory, use_opaque_style):
     import sys
     is_pypy = "__pypy__" in sys.builtin_module_names
 
     ctx = ctx_factory()
     queue = cl.CommandQueue(ctx)
 
-    _xfail_if_pocl_gpu(queue.device, "SVM")
-
     dev = ctx.devices[0]
 
     from pyopencl.characterize import has_coarse_grain_buffer_svm
@@ -1023,16 +1135,32 @@ def test_coarse_grain_svm(ctx_factory):
     if ("AMD" in dev.platform.name
             and dev.type & cl.device_type.GPU):
         pytest.xfail("AMD GPU crashes on SVM unmap")
+    if (dev.platform.vendor == "The pocl project"
+            and dev.type & cl.device_type.GPU
+            and "k40" in dev.name.lower()):
+        pytest.xfail("Crashes on K40s via POCL-CUDA")
 
+    dtype = np.dtype(np.float32)
     n = 3000
-    svm_ary = cl.SVM(cl.csvm_empty(ctx, (n,), np.float32, alignment=64))
-    if not is_pypy:
-        # https://bitbucket.org/pypy/numpy/issues/52
-        assert isinstance(svm_ary.mem.base, cl.SVMAllocation)
+    if use_opaque_style:
+        svm_ary = cl.SVMAllocation(ctx, n*dtype.itemsize, alignment=64,
+                                   flags=cl.svm_mem_flags.READ_WRITE)
+    else:
+        svm_ary = cl.SVM(cl.csvm_empty(ctx, (n,), dtype, alignment=64))
+        if not is_pypy:
+            # https://bitbucket.org/pypy/numpy/issues/52
+            assert isinstance(svm_ary.mem.base, cl.SVMAllocation)
 
-    cl.enqueue_svm_memfill(queue, svm_ary, np.zeros((), svm_ary.mem.dtype))
+    cl.enqueue_svm_memfill(queue, svm_ary, np.zeros((), dtype))
 
     with svm_ary.map_rw(queue) as ary:
+        if use_opaque_style:
+            ary = ary.view(dtype)
+        else:
+            assert ary is svm_ary.mem
+
+        assert ary.nbytes == n * dtype.itemsize
+
         ary.fill(17)
         orig_ary = ary.copy()
 
@@ -1043,21 +1171,28 @@ def test_coarse_grain_svm(ctx_factory):
         }
         """).build()
 
-    prg.twice(queue, svm_ary.mem.shape, None, svm_ary)
+    prg.twice(queue, (n,), None, svm_ary)
+
+    if dev.platform.vendor == "The pocl project" \
+            and dev.type & cl.device_type.GPU:
+        # clCreateBuffer from SVM doesn't work yet on GPU pocl
+        prg.twice(queue, (n,), None, svm_ary)
+    else:
+        prg.twice(queue, (n,), None, svm_ary.as_buffer(ctx))
 
     with svm_ary.map_ro(queue) as ary:
-        print(ary)
-        assert np.array_equal(orig_ary*2, ary)
+        if use_opaque_style:
+            ary = ary.view(dtype)
+        else:
+            assert ary is svm_ary.mem
+
+        assert np.array_equal(orig_ary*4, ary)
 
     new_ary = np.empty_like(orig_ary)
     new_ary.fill(-1)
 
-    if ctx.devices[0].platform.name != "Portable Computing Language":
-        # "Blocking memcpy is unimplemented (clEnqueueSVMMemcpy.c:61)"
-        # in pocl up to and including 1.0rc1.
-
-        cl.enqueue_copy(queue, new_ary, svm_ary)
-        assert np.array_equal(orig_ary*2, new_ary)
+    cl.enqueue_copy(queue, new_ary, svm_ary)
+    assert np.array_equal(orig_ary*4, new_ary)
 
     # {{{ https://github.com/inducer/pyopencl/issues/372
 
@@ -1083,6 +1218,10 @@ def test_coarse_grain_svm(ctx_factory):
 
     # }}}
 
+# }}}
+
+
+# {{{ test_fine_grain_svm
 
 def test_fine_grain_svm(ctx_factory):
     import sys
@@ -1121,10 +1260,14 @@ def test_fine_grain_svm(ctx_factory):
     print(ary)
     assert np.array_equal(orig_ary*2, ary)
 
+# }}}
+
+
+# {{{ test_map_dtype
 
 @pytest.mark.parametrize("dtype", [
     np.uint,
-    cltypes.uint2,
+    cltypes.uint2,      # type: ignore[attr-defined]
     ])
 def test_map_dtype(ctx_factory, dtype):
     if cl._PYPY:
@@ -1145,6 +1288,10 @@ def test_map_dtype(ctx_factory, dtype):
         print(array.dtype)
         assert array.dtype == dt
 
+# }}}
+
+
+# {{{ test_compile_link
 
 def test_compile_link(ctx_factory):
     ctx = ctx_factory()
@@ -1182,6 +1329,10 @@ def test_compile_link(ctx_factory):
     z.experiment(queue, (128**2,), (128,))
     queue.finish()
 
+# }}}
+
+
+# {{{ test_copy_buffer_rect
 
 def test_copy_buffer_rect(ctx_factory):
     ctx = ctx_factory()
@@ -1197,6 +1348,10 @@ def test_copy_buffer_rect(ctx_factory):
             src_origin=(0, 0), dst_origin=(1, 1),
             region=arr1.shape[::-1])
 
+# }}}
+
+
+# {{{ test_threaded_nanny_events
 
 def test_threaded_nanny_events(ctx_factory):
     # https://github.com/inducer/pyopencl/issues/296
@@ -1224,6 +1379,10 @@ def test_threaded_nanny_events(ctx_factory):
     t1.join()
     t2.join()
 
+# }}}
+
+
+# {{{ test_empty_ndrange
 
 @pytest.mark.parametrize("empty_shape", [(0,), (3, 0, 2)])
 def test_empty_ndrange(ctx_factory, empty_shape):
@@ -1244,6 +1403,10 @@ def test_empty_ndrange(ctx_factory, empty_shape):
 
     prg.add_two(queue, a.shape, None, a.data, allow_empty_ndrange=True)
 
+# }}}
+
+
+# {{{ test_command_queue_context_manager
 
 def test_command_queue_context_manager(ctx_factory):
     ctx = ctx_factory()
@@ -1253,14 +1416,107 @@ def test_command_queue_context_manager(ctx_factory):
     with pytest.warns(cl.CommandQueueUsedAfterExit):
         q.flush()
 
+# }}}
+
+
+# {{{ test_capture_call
+
+def test_capture_call(ctx_factory):
+    ctx = ctx_factory()
+    queue = cl.CommandQueue(ctx)
+
+    a_np = np.random.rand(500).astype(np.float32)
+    b_np = np.random.rand(500).astype(np.float32)
+
+    ctx = cl.create_some_context()
+    queue = cl.CommandQueue(ctx)
+
+    mf = cl.mem_flags
+    a_g = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a_np)
+    b_g = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=b_np)
+
+    prg = cl.Program(ctx, """
+    __kernel void sum(
+        __global const float *a_g, __global const float *b_g, __global float *res_g)
+    {
+    int gid = get_global_id(0);
+    res_g[gid] = a_g[gid] + b_g[gid];
+    }
+    """).build()
+
+    res_g = cl.Buffer(ctx, mf.WRITE_ONLY, a_np.nbytes)
+    from io import StringIO
+    sio = StringIO()
+    prg.sum.capture_call(sio, queue, a_np.shape, None, a_g, b_g, res_g)
+
+    compile_dict = {}
+    exec(compile(sio.getvalue(), "captured.py", "exec"), compile_dict)
+    compile_dict["main"]()
+
+# }}}
+
+
+# {{{ test_enqueue_copy_array
+
+def test_enqueue_copy_array(ctx_factory):
+    # https://github.com/inducer/pyopencl/issues/618
+    ctx = ctx_factory()
+    queue = cl.CommandQueue(ctx)
+
+    if ctx._get_cl_version() < (1, 2) or cl.get_cl_header_version() < (1, 2):
+        pytest.skip("requires CL 1.2")
+
+    if not queue.device.image_support:
+        pytest.skip("device has no image support")
+
+    image_format = cl.ImageFormat(cl.channel_order.RGBA, cl.channel_type.FLOAT)
+    flags = cl.mem_flags.READ_ONLY
+    image = np.ascontiguousarray(np.zeros((128, 128, 4), np.float32))
+    image_cl = cl.Image(ctx, flags, image_format,
+            shape=(image.shape[1], image.shape[0], 1), is_array=True)
+    cl.enqueue_copy(queue, dest=image, src=image_cl,
+            origin=(0, 0, 0), region=(image.shape[1], image.shape[0], 1))
+
+
+def test_enqueue_copy_array_2(ctx_factory):
+    # https://github.com/inducer/pyopencl/issues/618
+    ctx = ctx_factory()
+    queue = cl.CommandQueue(ctx)
+
+    if ctx._get_cl_version() < (1, 2) or cl.get_cl_header_version() < (1, 2):
+        pytest.skip("requires CL 1.2")
+
+    if not queue.device.image_support:
+        pytest.skip("device has no image support")
+
+    image_format = cl.ImageFormat(cl.channel_order.RGBA, cl.channel_type.FLOAT)
+    image = np.ascontiguousarray(np.zeros((128, 128, 4), np.float32))
+    image_shape = (image.shape[1], image.shape[0])
+    array_shape = (*image_shape, 1)
+    cl.Image(ctx, cl.mem_flags.READ_ONLY,
+                        image_format, shape=image_shape)
+    image_array_cl = cl.Image(ctx, cl.mem_flags.READ_ONLY,
+                              image_format, shape=array_shape, is_array=True)
+    image2_array_cl = cl.Image(ctx, cl.mem_flags.WRITE_ONLY,
+                               image_format, shape=array_shape, is_array=True)
+    buffer_cl = cl.Buffer(ctx, cl.mem_flags.WRITE_ONLY, size=image.nbytes)
+
+    cl._cl._enqueue_copy_image(
+            queue, src=image_array_cl, dest=image2_array_cl, src_origin=(0, 0, 0),
+            dest_origin=(0, 0, 0), region=array_shape)
+    cl._cl._enqueue_copy_image_to_buffer(
+            queue, src=image_array_cl, dest=buffer_cl, offset=0, origin=(0, 0, 0),
+            region=array_shape)
+
+# }}}
+
 
 if __name__ == "__main__":
-    # make sure that import failures get reported, instead of skipping the tests.
-    import pyopencl  # noqa
-
     import sys
     if len(sys.argv) > 1:
         exec(sys.argv[1])
     else:
         from pytest import main
         main([__file__])
+
+# vim: foldmethod=marker
